__version__ = "0.1.3"

__all__ = [
    "__version__",
    "gdf_apply",
    "xy_to_lonlat",
    "lonlat_to_xy",
    "set_uid",
    "read_gdf",
    "pre_proc_gdf",
    "clip_gdf",
    "is_categorical",
    "get_holes",
    "gdf_to_file",
]

from .apply import gdf_apply
from .merging.save_utils import gdf_to_file
from .utils import (
    clip_gdf,
    get_holes,
    is_categorical,
    lonlat_to_xy,
    pre_proc_gdf,
    read_gdf,
    set_uid,
    xy_to_lonlat,
)



import warnings
from pathlib import Path
from typing import Dict, Sequence, Tuple, Union

import geopandas as gpd
import pandas as pd
import shapely
from pandas.api.types import (
    is_bool_dtype,
    is_categorical_dtype,
    is_object_dtype,
    is_string_dtype,
)
from pyproj import CRS, Transformer
from shapely.geometry import Polygon

from .apply import gdf_apply

__all__ = [
    "set_uid",
    "read_gdf",
    "pre_proc_gdf",
    "clip_gdf",
    "is_categorical",
    "get_holes",
    "xy_to_lonlat",
    "lonlat_to_xy",
]


def is_categorical(col: pd.Series) -> bool:
    """Check if a column is categorical."""
    return (
        is_categorical_dtype(col)
        or is_string_dtype(col)
        or is_object_dtype(col)
        or is_bool_dtype(col)
    )


def set_uid(
    gdf: gpd.GeoDataFrame, start_ix: int = 0, id_col: str = "uid", drop: bool = False
) -> gpd.GeoDataFrame:
    """Set a unique identifier column to gdf.

    Note:
        by default sets a running index column to gdf as the uid.

    Parameters:
        gdf (gpd.GeoDataFrame):
            Input Geodataframe.
        start_ix (int, default=0):
            The starting index of the id column.
        id_col (str, default="uid"):
            The name of the column that will be used or set to the id.
        drop (bool, default=False):
            Drop the column after it is added to index.

    Returns:
        gpd.GeoDataFrame:
            The input gdf with a "uid" column added to it.

    Examples:
        >>> from cellseg_gsontools import set_uid
        >>> gdf = set_uid(gdf, drop=True)
    """
    gdf = gdf.copy()
    if id_col not in gdf.columns:
        gdf[id_col] = range(start_ix, len(gdf) + start_ix)

    gdf = gdf.set_index(id_col, drop=drop)

    return gdf


def _set_cols(property_dict: Dict):
    return property_dict["properties"]


def _set_geom(property_dict: Dict):
    return property_dict["geometry"]


def _set_gdf(gdf):
    gdf["geometry"] = gdf_apply(gdf, _set_geom, columns=["features"], parallel=False)
    gdf["properties"] = gdf_apply(gdf, _set_cols, columns=["features"], parallel=False)
    return gdf


def _get_class(property_dict: Dict) -> str:
    """Return the class of the gdf."""
    if "classification" in property_dict.keys():
        return property_dict["classification"]["name"]
    return property_dict["name"]


def _get_prob(property_dict: Dict) -> str:
    """Return the class probabilities of the gdf."""
    if "classification" in property_dict.keys():
        return property_dict["classification"]["probabilities"]
    return property_dict["probabilities"]


def read_gdf(
    fname: Union[Path, str],
    silence_warnigns: bool = True,
) -> gpd.GeoDataFrame:
    """Read a file into a geodataframe.

    This is a wrapper around `geopandas` I/O that adds some extra
    functionality.

    Note:
        Allowed formats:

        - `.json`,
        - `.geojson`,
        - `.feather`,
        - `.parquet`

    Parameters:
        fname (Union[Path, str]):
            The filename of the gson file.
        silence_warnigns (bool):
            Whether to silence warnings, by default True.

    Raises:
        ValueError:
            If suffix is not one of ".json", ".geojson", ".feather", ".parquet".

    Returns:
        gpd.GeoDataFrame:
            The geodataframe.

    Examples:
        Read a geojson file that is QuPath-readable.
        >>> from cellseg_gsontools.utils import read_gdf
        >>> gdf = read_gdf("path/to/file.json")
    """
    fname = Path(fname)
    format = fname.suffix
    allowed_formats = (".json", ".geojson", ".feather", ".parquet")
    if format not in allowed_formats:
        raise ValueError(
            f"Illegal `format`. Got: {format}. Allowed: {allowed_formats}."
        )

    if format == ".json":
        df = pd.read_json(fname)
    elif format == ".geojson":
        try:
            df = gpd.read_file(fname)
        except Exception:
            df = pd.read_json(fname)
            df = _set_gdf(df)
    elif format == ".feather":
        df = gpd.read_feather(fname)
    elif format == ".parquet":
        df = gpd.read_parquet(fname)

    if df.empty:
        if not silence_warnigns:
            warnings.warn(f"Empty geojson file: {fname.name}. Returning empty gdf.")
        return df

    property_col = "properties" if "properties" in df.columns else "classification"

    if "class_name" not in df.columns:
        try:
            df["class_name"] = gdf_apply(df, _get_class, columns=[property_col])
        except KeyError:
            if not silence_warnigns:
                warnings.warn(
                    f"Could not find 'name' key in {property_col} column."
                    "Can't set the `class_name` column to the output gdf."
                )

    if "class_probs" not in df.columns:
        try:
            df["class_probs"] = gdf_apply(df, _get_prob, columns=[property_col])
        except KeyError:
            if not silence_warnigns:
                warnings.warn(
                    f"Could not find 'probabilities' key in {property_col} column. "
                    "Can't set the `class_probs` column to the output gdf."
                )

    df["geometry"] = gdf_apply(df, shapely.geometry.shape, columns=["geometry"])
    return gpd.GeoDataFrame(df).set_geometry("geometry")


def pre_proc_gdf(
    gdf: gpd.GeoDataFrame, min_size: int = None
) -> Union[gpd.GeoDataFrame, None]:
    """Apply some light pre-processing of a geodataframe.

    Namely, remove invalid polygons, empty geometries and add bounds to the gdf.

    Parameters:
        gdf (gpd.GeoDataFrame):
            Input geodataframe.
        min_size (int, optional):
            The minimum size of the polygons in pixels.

    Returns:
        gpd.GeoDataFrame:
            The pre-processed gdf or None if input gdf is empty or None.

    Examples:
        >>> from cellseg_gsontools import pre_proc_gdf
        >>> gdf = pre_proc_gdf(gdf, min_size=100)
    """
    if gdf.empty or gdf is None:
        return gdf

    # drop invalid geometries if there are any after buffer
    gdf.geometry = gdf.geometry.buffer(0)
    gdf = gdf[gdf.is_valid]

    # drop empty geometries
    gdf = gdf[~gdf.is_empty]

    # if there are multipolygon geometries, explode them
    if "MultiPolygon" in list(gdf["geometry"].geom_type):
        gdf = gdf.explode(index_parts=False).reset_index(drop=True)

    # drop geometries that are less than min_size pixels
    if min_size is not None:
        gdf = gdf[gdf.area > min_size]

    # drop geometries that are not polygons
    gdf = gdf[gdf.geom_type == "Polygon"]

    try:
        # add bounding box coords of the polygons to the gdfs
        # and correct for the max coords
        gdf["xmin"] = gdf.bounds["minx"].astype(int)
        gdf["ymin"] = gdf.bounds["miny"].astype(int)
        gdf["ymax"] = gdf.bounds["maxy"].astype(int) + 1
        gdf["xmax"] = gdf.bounds["maxx"].astype(int) + 1
    except Exception:
        warnings.warn("Could not create bounds cols to gdf.", RuntimeWarning)

    return gdf


def clip_gdf(
    gdf: gpd.GeoDataFrame, bbox: Tuple[int, int, int, int]
) -> gpd.GeoDataFrame:
    """Clip a gdf to a bounding box.

    Parameters:
        gdf (gpd.GeoDataFrame):
            Input geodataframe.
        bbox (Tuple[int, int, int, int]):
            Bounding box to clip to. Format: (xmin, ymin, xmax, ymax).

    Returns:
        gpd.GeoDataFrame:
            The Clipped gdf.

    Examples:
        >>> from cellseg_gsontools import clip_gdf
        >>> gdf = clip_gdf(gdf, bbox=(0, 0, 100, 100))
    """
    xmin = bbox[0]
    ymin = bbox[1]
    xmax = bbox[2]
    ymax = bbox[3]

    crop = Polygon(
        [(xmin, ymin), (xmin, ymax), (xmax, ymax), (xmax, ymin), (xmin, ymin)]
    )

    return gdf.clip(crop)


def xy_to_lonlat(
    x: Union[float, Sequence], y: Union[float, Sequence]
) -> Tuple[Union[float, Sequence], Union[float, Sequence]]:
    """Converts x, y coordinates to lon, lat coordinates.

    Parameters:
        x (Union[float, Sequence]):
            x coordinate(s).
        y (Union[float, Sequence]):
            y coordinate(s).

    Returns:
        Tuple[Union[float, Sequence], Union[float, Sequence]]:
            lon, lat coordinates.

    Examples:
        >>> from cellseg_gsontools import xy_to_lonlat
        >>> from shapely.geometry import Polygon
        >>> poly = Polygon([(0, 0), (0, 1), (1, 1)])
        >>> x, y = poly.exterior.coords.xy
        >>> lon, lat = xy_to_lonlat(x, y)
        >>> lon, lat
        (array('d', [10.511, 10.511, 10.511, 10.511]),
         array('d', [0.0, 9.019e-06, 9.019e-06, 0.0]))
    """
    crs_utm = CRS(proj="utm", zone=33, ellps="WGS84")
    crs_latlon = CRS(proj="latlong", zone=33, ellps="WGS84")
    transformer = Transformer.from_crs(crs_utm, crs_latlon, always_xy=True)
    lonlat = transformer.transform(x, y)

    return lonlat[0], lonlat[1]


def lonlat_to_xy(
    lon: Union[float, Sequence], lat: Union[float, Sequence]
) -> Tuple[Union[float, Sequence], Union[float, Sequence]]:
    """Converts lon, lat coordinates to x, y coordinates.

    Parameters:
        lon (Union[float, Sequence]):
            Longitude coordinate(s).
        lat (Union[float, Sequence]):
            Latitude coordinate(s).

    Returns:
        Tuple[Union[float, Sequence], Union[float, Sequence]]:
            x, y coordinates.

    Examples:
        >>> from shapely.geometry import Polygon
        >>> from cellseg_gsontools import lonlat_to_xy
        >>> poly = Polygon([(10, 10), (10, 0), (20, 10)])
        >>> lon, lat = poly.exterior.coords.xy
        >>> x, y = lonlat_to_xy(lon, lat)
        >>> x, y
        (array('d', [-48636.648, -57087.120, 1048636.648, -48636.648]),
         array('d', [1109577.311, 0.0, 1109577.311, 1109577.311]))
    """
    crs_utm = CRS(proj="utm", zone=33, ellps="WGS84")
    crs_latlon = CRS(proj="latlong", zone=33, ellps="WGS84")
    transformer = Transformer.from_crs(crs_latlon, crs_utm, always_xy=True)
    xy = transformer.transform(lon, lat)

    return xy[0], xy[1]


def get_holes(poly: Polygon, to_lonlat: bool = True) -> Sequence[Sequence[float]]:
    """Get holes from a shapely Polygon.

    Parameters:
        poly (Polygon):
            Polygon to get holes from.
        to_lonlat (bool, optional):
            Whether to convert to lonlat coordinates, by default True.

    Returns:
        Sequence[Sequence[float]]:
            A list of xy coordinate tuples.

    Examples:
        >>> from cellseg_gsontools import get_holes
        >>> holes = get_holes(poly)
    """
    holes = []
    for interior in poly.interiors:
        x, y = interior.coords.xy
        if to_lonlat:
            x, y = xy_to_lonlat(x, y)
        holes.append(list(zip(x, y)))

    return holes



import re
from pathlib import Path
from typing import List, Tuple, Union

import geopandas as gpd

from ..apply import gdf_apply

__all__ = [
    "check_format",
    "get_xy_coords",
    "get_file_from_coords",
    "gdf_to_file",
]


def _add_qupath_classification(row: gpd.GeoSeries) -> gpd.GeoSeries:
    """Add QuPath properties to a row of a gdf."""
    c = row["class_name"]
    props = {"name": c, "color": None}
    return props


def gdf_to_file(
    gdf: gpd.GeoDataFrame,
    out_fn: Union[str, Path],
    format: str = ".feather",
) -> None:
    """Write a geojson/feather/parquet file from a gdf.

    This is wrapper around `geopandas.GeoDataFrame` I/O methods
    that adds some extra functionality.

    Parameters:
        gdf (gpd.GeoDataFrame):
            The input gdf.
        out_fn (Union[str, Path]):
            The output filename.
        format (str):
            The output format. One of ".feather", ".parquet", ".geojson".

    Raises:
        ValueError: If `format` is not one of ".feather", ".geojson", ".parquet".
        ValueError: If the input gdf does not have a "class_name" column.

    Examples:
        Write a geojson file.
        >>> from cellseg_gsontools import gdf_to_file
        >>> gdf_to_file(gdf, "out.geojson")
    """
    out_fn = Path(out_fn)
    if format not in (".feather", ".parquet", ".geojson", None):
        raise ValueError(
            f"Invalid format. Got: {format}. Allowed: .feather, .parquet, .geojson"
        )

    if "class_name" not in gdf.columns:
        raise ValueError("The input gdf needs to have a 'class_name' column.")

    # add objectType col (QuPath)
    if "objectType" not in gdf.columns:
        gdf["objectType"] = "annotation"

    # add classification col (QuPath)
    if "classification" not in gdf.columns:
        gdf["classification"] = gdf_apply(
            gdf, _add_qupath_classification, col=None, axis=1
        )

    if format == ".feather":
        gdf.to_feather(out_fn.with_suffix(".feather"))
    elif format == ".parquet":
        gdf.to_parquet(out_fn.with_suffix(".parquet"))
    elif format == ".geojson":
        gdf.to_file(out_fn.with_suffix(".geojson"), driver="GeoJSON", index=False)


def check_format(fname: Union[Path, str]) -> None:
    """Check if the input file has the correct format.

    Parameters
        fname (str):
            The filename.

    Raises:
        ValueError: If not all coordinates were found in filename.
        ValueError: If both x and y coordinates are not present in filename.
    """
    fn = Path(fname)
    if fn.suffix not in (".json", ".geojson", ".png", ".feather", ".parquet"):
        raise ValueError(
            f"Input file {fn} has wrong format. Expected '.png', '.json' or '.geojson'."
        )

    has_x = False
    has_y = False

    # get the x and y coordinates from the filename
    # NOTE: fname needs to contain x & y-coordinates in x_[coord1]_y_[coord2]-format
    # or x-[coord1]_y-[coord2]-format. The order of x and y can be any.
    xy_str: List[str] = re.findall(r"x\d+|y\d+|x_\d+|y_\d+|x-\d+|y-\d+", fn.as_posix())

    try:
        for s in xy_str:
            if "x" in s:
                has_x = True
            elif "y" in s:
                has_y = True
    except IndexError:
        raise ValueError(
            "Not all coordinates were found in filename. "
            f"Filename has to be in 'x-[coord1]_y-[coord2]'-format, Got: {fn.name}"
        )

    if not has_x or not has_y:
        raise ValueError(
            "Both x and y coordinates have to be present in filename. "
            f"Got: {xy_str}. Filename has to be in 'x-[coord1]_y-[coord2]'-format."
        )

    return


def get_xy_coords(fname: Union[Path, str]) -> Tuple[int, int]:
    """Get the x and y-coordinates from a filename.

    Note:
        The filename needs to contain x & y-coordinates in
        "x-[coord1]_y-[coord2]"-format

    Parameters:
        fname (str):
            The filename. Has to contain x & y-coordinates

    Raises:
        ValueError: If not the delimeter of x and y- coord is not '_' or '-'.

    Returns:
        Tuple[int, int]: The x and y-coordinates in this order.
    """
    check_format(fname)

    if isinstance(fname, Path):
        fname = fname.as_posix()

    xy_str: List[str] = re.findall(r"x\d+|y\d+|x_\d+|y_\d+|x-\d+|y-\d+", fname)
    xy: List[int] = [0, 0]
    for s in xy_str:
        if "x" in s:
            if "_" in s:
                xy[0] = int(s.split("_")[1])
            elif "-" in s:
                xy[0] = int(s.split("-")[1])
            elif "x" in s and "-" not in s and "_" not in s:
                xy[0] = int(s.split("x")[1])
            else:
                raise ValueError(
                    "The fname needs to contain x & y-coordinates in "
                    f"'x-[coord1]_y-[coord2]'-format. Got: {fname}"
                )
        elif "y" in s:
            if "_" in s:
                xy[1] = int(s.split("_")[1])
            elif "-" in s:
                xy[1] = int(s.split("-")[1])
            elif "y" in s and "-" not in s and "_" not in s:
                xy[1] = int(s.split("y")[1])
            else:
                raise ValueError(
                    "The fname needs to contain x & y-coordinates in "
                    f"'x-[coord1]_y-[coord2]'-format. Got: {fname}"
                )

    return xy[0], xy[1]


def get_file_from_coords(
    files: List[Union[Path, str]], x: int, y: int
) -> Union[str, None]:
    """Get the correct file name from a list of filenames given x and y-coords.

    Note:
        fname needs to contain x & y-coordinates in  "x-[coord1]_y-[coord2]"-format.

    Parameters:
        files (List[Union[Path, str]]): A list of paths to the files.
        x (int): x-coord in pixels.
        y (int): y-coord in pixels.

    Returns:
        str or None: returns the file name if it exists in the given input dir, else None.
    """
    for f in files:
        if get_xy_coords(f) == (x, y):
            return f

    return



from typing import Tuple, Union

import numpy as np
from shapely.geometry import Point, Polygon

__all__ = ["axis_len", "axis_angle", "_dist", "_azimuth"]


def _dist(p1: Point, p2: Point) -> float:
    """Compute distance between two points."""
    return p1.distance(p2)


def _azimuth(p1: Point, p2: Point) -> float:
    """Azimuth between 2 points (interval 0 - 180)."""
    angle = np.arctan2(p2[0] - p1[0], p2[1] - p1[1])
    return np.degrees(angle) if angle > 0 else np.degrees(angle) + 180


def axis_len(
    mrr: Polygon, which: str = "both", **kwargs
) -> Union[float, Tuple[float, float]]:
    """Compute major and minor axis from minimum rotated rectangle.

    Parameters
    ----------
        mrr : Polygon
            Input minimum rotated rectangle of a shapely polygon object.
        which : str, default="both"
            One of ("both", "major", "minor").

    Returns
    -------
        Union[float, Tuple[float, float]]
            The major and/or minor axis lengths.
    """
    allowed = ("both", "major", "minor")
    if which not in allowed:
        raise ValueError(f"Illegal arg `which`. Got: {which}. Allowed: {allowed}")

    axis1 = _dist(Point(mrr[0]), Point(mrr[3]))
    axis2 = _dist(Point(mrr[0]), Point(mrr[1]))

    minoraxis_len = min([axis1, axis2])
    majoraxis_len = max([axis1, axis2])

    if which == "major":
        return majoraxis_len
    elif which == "minor":
        return minoraxis_len
    else:
        return majoraxis_len, minoraxis_len


def axis_angle(
    mrr: Polygon, which: str = "both", **kwargs
) -> Union[float, Tuple[float, float]]:
    """Compute major and minor axis angles from minimum rotated rectangle.

    Parameters
    ----------
        polygon : Polygon
            Input minimum rotated rectangle polygon object.
        which : str, default="both"
            One of ("both", "major", "minor").

    Returns
    -------
        Union[float, Tuple[float, float]]
            The major and/or minor axis angles in degrees.
    """
    allowed = ("both", "major", "minor")
    if which not in allowed:
        raise ValueError(f"Illegal arg `which`. Got: {which}. Allowed: {allowed}")

    axis1 = _dist(Point(mrr[0]), Point(mrr[3]))
    axis2 = _dist(Point(mrr[0]), Point(mrr[1]))

    if axis1 <= axis2:
        minoraxis_ang = _azimuth(mrr[0], mrr[3])
        majoraxis_ang = _azimuth(mrr[0], mrr[1])
    else:
        minoraxis_ang = _azimuth(mrr[0], mrr[1])
        majoraxis_ang = _azimuth(mrr[0], mrr[3])

    if which == "major":
        return majoraxis_ang
    elif which == "minor":
        return minoraxis_ang
    else:
        return majoraxis_ang, minoraxis_ang



from typing import List, Sequence, Union

import mapclassify
import numpy as np
import pandas as pd
from libpysal.weights import W

from .geometry.axis import _dist
from .utils import is_categorical

__all__ = [
    "neighborhood",
    "nhood_counts",
    "nhood_vals",
    "nhood_type_count",
    "nhood_dists",
]


def neighborhood(
    node: Union[int, pd.Series],
    spatial_weights: W,
    include_self: bool = True,
    ret_n_neighbors: bool = False,
) -> Union[List[int], int]:
    """Get immediate neighborhood of a node given the spatial weights obj.

    Note:
        This function is designed to be used with the `gdf_apply` function.
        See the example.

    Note:
        The neighborhood contains the given node itself by default.

    Parameters:
        node (int or pd.Series):
            Input node uid.
        spatial_weights (libysal.weights.W):
            Libpysal spatial weights object.
        include_self (bool):
            Flag, whether to include the node itself in the neighborhood.
            Defaults to True.
        ret_n_neighbors (bool):
            If True, instead of returning a sequence of the neighbor node uids
            returns just the number of neighbors. Defaults to False.

    Returns:
        List[int] or int:
            A list of the neighboring node uids. E.g. [1, 4, 19].
            or the number of neighbors if `ret_n_neighbors=True`.

    Examples:
        Use `gdf_apply` to extract the neighboring nodes for each node/cell
        >>> from functools import partial
        >>> from cellseg_gsontools.data import gland_cells
        >>> from cellseg_gsontools.graphs import fit_graph
        >>> from cellseg_gsontools.utils import set_uid
        >>> from cellseg_gsontools.apply import gdf_apply
        >>> from cellseg_gsontools.neighbors import neighborhood
        >>> gc = gland_cells()
        >>> # To fit the delaunay graph, we need to set a unique id for each cell first
        >>> gc = set_uid(gc, id_col="uid")
        >>> w = fit_graph(gc, type="delaunay", thresh=100, id_col="uid")
        >>> # Get the neihgboring nodes of the graph
        >>> func = partial(neighborhood, spatial_weights=w)
        >>> gc["nhood"] = gdf_apply(gc, func, columns=["uid"])
        >>> gc["nhood"].head(5)
                uid
        0       [0, 1, 3, 4, 483, 484]
        1     [1, 0, 4, 482, 483, 487]
        2       [2, 3, 5, 6, 484, 493]
        3      [3, 0, 2, 4, 5, 7, 484]
        4    [4, 0, 1, 3, 7, 487, 488]
        Name: nhood, dtype: object
    """
    if isinstance(node, pd.Series):
        node = node.iloc[0]  # assume that the series is a row

    nhood = np.nan
    if ret_n_neighbors:
        nhood = spatial_weights.cardinalities[node]
    elif node in spatial_weights.neighbors.keys():
        # get spatial neighborhood
        nhood = spatial_weights.neighbors[node]
        if include_self:
            nhood = [node] + list(nhood)

    return nhood


def nhood_vals(nhood: Sequence[int], values: pd.Series, **kwargs) -> np.ndarray:
    """Get the values of objects in the neighboring nodes.

    Note:
        This function is designed to be used with the `gdf_apply` function.
        See the example.

    Parameters:
        nhood (Sequence[int]):
            A list or array of neighboring node uids.
        values (pd.Series):
            A value column-vector of shape (N, ).
        **kwargs (Dict[str, Any]):
            Additional keyword arguments. Not used.

    Returns:
        np.ndarray:
            The counts vector of the given values vector. Shape (n_classes, )

    Examples:
        Use `gdf_apply` to get the neighborhood values for each area of a cell
        >>> from functools import partial
        >>> from cellseg_gsontools.data import gland_cells
        >>> from cellseg_gsontools.graphs import fit_graph
        >>> from cellseg_gsontools.utils import set_uid
        >>> from cellseg_gsontools.apply import gdf_apply
        >>> from cellseg_gsontools.neighbors import neighborhood, nhood_vals
        >>> gc = gland_cells()
        >>> # To fit the delaunay graph, we need to set a unique id for each cell first
        >>> gc = set_uid(gc, id_col="uid")
        >>> w = fit_graph(gc, type="delaunay", thresh=100, id_col="uid")
        >>> # Get the neihgboring nodes of the graph
        >>> func = partial(neighborhood, spatial_weights=w)
        >>> gc["nhood"] = gdf_apply(gc, func, columns=["uid"])
        >>> # get the area values of the neighbors
        >>> func = partial(nhood_vals, values=gc.area.round(2))
        >>> gc["neighbor_areas"] = gdf_apply(
        ...     gc,
        ...     func=func,
        ...     parallel=True,
        ...     columns=["nhood"],
        ... )
        >>> gc["neighbor_areas"].head(5)
            uid
        0     [520.24, 565.58, 435.91, 302.26, 241.85, 418.02]
        1     [565.58, 520.24, 302.26, 318.15, 241.85, 485.71]
        2      [721.5, 435.91, 556.05, 466.96, 418.02, 678.35]
        3    [435.91, 520.24, 721.5, 302.26, 556.05, 655.42...
        4    [302.26, 520.24, 565.58, 435.91, 655.42, 485.7...
        Name: neighbor_areas, dtype: object
    """
    if isinstance(nhood, pd.Series):
        nhood = nhood.iloc[0]  # assume that the series is a row

    nhood_vals = np.array([0])
    if nhood not in (None, np.nan) and isinstance(nhood, (Sequence, np.ndarray)):
        nhood_vals = values.loc[nhood].to_numpy()

    return nhood_vals


def nhood_counts(
    nhood: Sequence[int], values: pd.Series, bins: Sequence, **kwargs
) -> np.ndarray:
    """Get the counts of objects that belong to bins/classes in the neighborhood.

    Note:
        This function is designed to be used with the `gdf_apply` function.
        See the example.

    Parameters:
        nhood (Sequence[int]):
            A list or array of neighboring node uids.
        values (pd.Series):
            A value column-vector of shape (N, ).
        bins (Sequence):
            The bins of any value vector. Shape (n_bins, ).
        return_vals (bool, optional):
            If True, also, the values the values are. Defaults to False.
        **kwargs (Dict[str, Any]):
            Additional keyword arguments. Not used.

    Returns:
        np.ndarray:
            The counts vector of the given values vector. Shape (n_classes, )

    Examples:
        Use `gdf_apply` to compute the neighborhood counts for each areal bin
        >>> import mapclassify
        >>> from functools import partial
        >>> from cellseg_gsontools.data import gland_cells
        >>> from cellseg_gsontools.graphs import fit_graph
        >>> from cellseg_gsontools.utils import set_uid
        >>> from cellseg_gsontools.apply import gdf_apply
        >>> from cellseg_gsontools.neighbors import neighborhood, nhood_vals, nhood_counts
        >>> gc = gland_cells()
        >>> # To fit the delaunay graph, we need to set a unique id for each cell first
        >>> gc = set_uid(gc, id_col="uid")
        >>> w = fit_graph(gc, type="delaunay", thresh=100, id_col="uid")
        >>> # Get the neihgboring nodes of the graph
        >>> func = partial(neighborhood, spatial_weights=w)
        >>> gc["nhood"] = gdf_apply(gc, func, columns=["uid"])
        >>> # get the area values of the neighbors
        >>> func = partial(nhood_vals, values=gc.area.round(2))
        >>> gc["neighbor_areas"] = gdf_apply(
        ...     gc,
        ...     func=func,
        ...     parallel=True,
        ...     columns=["nhood"],
        ... )
        >>> bins = mapclassify.Quantiles(gc.area, k=5)
        >>> func = partial(nhood_counts, values=gc.area, bins=bins.bins)
        >>> gc["area_bins"] = gdf_apply(
        ...     gc,
        ...     func,
        ...     columns=["nhood"],
        ... )
        >>> gc["area_bins"].head(5)
        uid
        0    [0, 2, 0, 3, 1]
        1    [0, 2, 1, 2, 1]
        2    [0, 0, 0, 3, 3]
        3    [0, 1, 0, 3, 3]
        4    [0, 1, 0, 3, 3]
        Name: area_bins, dtype: object
    """
    if isinstance(nhood, pd.Series):
        nhood = nhood.iloc[0]  # assume that the series is a row

    counts = np.array([0])
    if nhood not in (None, np.nan) and isinstance(nhood, (Sequence, np.ndarray)):
        nhood_vals = values.loc[nhood]

        if is_categorical(nhood_vals):
            counts = nhood_vals.value_counts().values
        else:
            sample_bins = mapclassify.UserDefined(nhood_vals, bins)
            counts = sample_bins.counts

    return counts


def nhood_type_count(
    cls_neighbors: Sequence, cls: Union[int, str], frac: bool = True, **kwargs
) -> float:
    """Get the number of nodes of a specific category in a neighborhood of a node.

    Note:
        This function is designed to be used with the `gdf_apply` function.
        See the example.

    Parameters:
        cls_neihbors (Sequence):
            A array/list (int or str) containing a category for each value in the data.
        cls (int or str):
            The specific category.
        frac (bool, optional):
            Flag, whether to return the fraction instead of the count. Defaults to True.
        **kwargs (Dict[str, Any])]):
            Additional keyword arguments. Not used.
    Returns:
        float:
            The count or fraction of a node of specific category in a neighborhood.

    Raises:
        TypeError:
            If the `cls_neighbors` is not categorical.

    Examples:
        Use `gdf_apply` to get the neighborhood fractions of immune cells for each node
        >>> from functools import partial
        >>> from cellseg_gsontools.data import gland_cells
        >>> from cellseg_gsontools.graphs import fit_graph
        >>> from cellseg_gsontools.utils import set_uid
        >>> from cellseg_gsontools.apply import gdf_apply
        >>> from cellseg_gsontools.neighbors import (
        ...     neighborhood,
        ...     nhood_vals,
        ...     hnood_type_count,
        ... )
        >>> gc = gland_cells()
        >>> # To fit the delaunay graph, we need to set a unique id for each cell first
        >>> gc = set_uid(gc, id_col="uid")
        >>> w = fit_graph(gc, type="delaunay", thresh=100, id_col="uid")
        >>> # Get the neihgboring nodes of the graph
        >>> func = partial(neighborhood, spatial_weights=w)
        >>> gc["nhood"] = gdf_apply(gc, func, columns=["uid"])
        >>> # get the classes of the neighbors
        >>> func = partial(nhood_vals, values=gc.class_name)
        >>> gc["neighbor_classes"] = gdf_apply(
        ...     gc,
        ...     func=func,
        ...     parallel=True,
        ...     columns=["nhood"],
        ... )
        >>> func = partial(nhood_type_count, cls="inflammatory", frac=True)
        >>> gc["n_immune_neighbors"] = gdf_apply(
        ...    gc,
        ...    func=func,
        ...    parallel=True,
        ...    columns=["neighbor_classes"],
        >>> )
        >>> gc[gc["n_immune_neighbors"] > 0]["n_immune_neighbors"].head(5)
        uid
        39    0.333333
        40    0.111111
        42    0.166667
        44    0.375000
        48    0.166667
        Name: n_immune_neighbors, dtype: float64
    """
    if isinstance(cls_neighbors, pd.Series):
        cls_neighbors = cls_neighbors.iloc[0]  # assume that the series is a row

    ret = 0
    if isinstance(cls_neighbors, (Sequence, np.ndarray)):
        if len(cls_neighbors) > 0:
            if not isinstance(cls_neighbors[0], (int, str)):
                raise TypeError("cls_neighbors must contain int of str values.")

        t, c = np.unique(cls_neighbors, return_counts=True)

        ret = 0.0
        if cls in t:
            ix = np.where(t == cls)
            ret = c[ix][0]
            if frac:
                ret = ret / np.sum(c)

    return ret


def nhood_dists(
    nhood: Sequence[int],
    centroids: pd.Series,
    ids: pd.Series = None,
    invert: bool = False,
) -> np.ndarray:
    """Compute the neighborhood distances between the center node.

    Note:
        This function is designed to be used with the `gdf_apply` function.
        See the example.

    Note:
        It is assumed that the center node is the first index in the `nhood`
        array. Use `include_self=True` in `neighborhood` to include the center.

    Parameters:
        nhood (Sequence[int]):
            An array containing neighbor indices. The first index is assumed to be
            the center node.
        centroids (pd.Series):
            A pd.Series array containing the centroid Points of the full gdf.
        ids (pd.Series):
            A pd.Series array containing the ids of the full gdf.
        invert (bool):
            Flag, whether to invert the distances. E.g. 1/dists. Defaults to False.

    Returns:
        np.ndarray:
            An array containing the distances between the center node and its
            neighborhood.

    Examples:
        Use `gdf_apply` to extract the neighboring node distances for each node/cell
        >>> from functools import partial
        >>> from cellseg_gsontools.data import gland_cells
        >>> from cellseg_gsontools.graphs import fit_graph
        >>> from cellseg_gsontools.utils import set_uid
        >>> from cellseg_gsontools.apply import gdf_apply
        >>> from cellseg_gsontools.neighbors import neighborhood, nhood_dists
        >>> gc = gland_cells()
        >>> # To fit the delaunay graph, we need to set a unique id for each cell first
        >>> gc = set_uid(gc, id_col="uid")
        >>> w = fit_graph(gc, type="delaunay", thresh=100, id_col="uid")
        >>> # Get the neihgboring nodes of the graph
        >>> func = partial(neighborhood, spatial_weights=w)
        >>> gc["nhood"] = gdf_apply(gc, func, columns=["uid"])
        >>> func = partial(nhood_dists, centroids=gc.centroid)
        >>> gc["nhood_dists"] = gdf_apply(
        ...     gc,
        ...     func,
        ...     columns=["nhood"],
        ... )
        >>> gc["nhood_dists"].head(5)
        uid
        0        [0.0, 26.675, 24.786, 30.068, 30.228, 41.284]
        1        [0.0, 26.675, 23.428, 42.962, 39.039, 23.949]
        2        [0.0, 25.577, 39.348, 46.097, 34.309, 29.478]
        3    [0.0, 24.786, 25.577, 39.574, 37.829, 47.16, 3...
        4    [0.0, 30.068, 23.428, 39.574, 29.225, 36.337, ...
        Name: nhood_dists, dtype: object
    """
    if isinstance(nhood, pd.Series):
        nhood = nhood.iloc[0]  # assume that the series is a row

    nhood_dists = np.array([0])
    if nhood not in (None, np.nan) and isinstance(nhood, (Sequence, np.ndarray)):
        if ids is not None:
            nhood = ids[ids.isin(nhood)].index

        node = nhood[0]
        center_node = centroids.loc[node]
        nhood_nodes = centroids.loc[nhood].to_numpy()
        nhood_dists = np.array(
            [np.round(_dist(center_node, c), 3) for c in nhood_nodes]
        )
        if invert:
            nhood_dists = np.round(np.reciprocal(nhood_dists, where=nhood_dists > 0), 3)

    return nhood_dists



from typing import Tuple

import geopandas as gpd
import numpy as np
import shapely
from shapely.geometry import Polygon

from ..apply import gdf_apply
from .axis import axis_angle, axis_len
from .circle import inscribing_circle

__all__ = [
    "major_axis_len",
    "minor_axis_len",
    "major_axis_angle",
    "minor_axis_angle",
    "compactness",
    "circularity",
    "convexity",
    "solidity",
    "elongation",
    "eccentricity",
    "fractal_dimension",
    "sphericity",
    "shape_index",
    "rectangularity",
    "squareness",
    "equivalent_rectangular_index",
    "shape_metric",
    "SHAPE_LOOKUP",
]


def major_axis_len(polygon: Polygon) -> float:
    """Compute the major axis length of a polygon.

    Note:
        The major axis is the (x,y) endpoints of the longest line that
        can be drawn through the object. Major axis length is the pixel
        distance between the major-axis endpoints
        - [Wirth](http://www.cyto.purdue.edu/cdroms/micro2/content/education/wirth10.pdf)

    Parameters:
        polygon (Polygon):
            Input shapely polygon object.

    Returns:
        float:
            The length of the major axis.
    """
    mrr = polygon.minimum_rotated_rectangle.exterior.coords
    return axis_len(mrr, "major")


def minor_axis_len(polygon: Polygon) -> float:
    """Compute the minor axis length of a polygon.

    Note:
        The minor axis is the (x,y) endpoints of the longest line that
        can be drawn through the object whilst remaining perpendicular
        with the major-axis. Minor axis length is the pixel distance
        between the minor-axis endpoints
        - [Wirth](http://www.cyto.purdue.edu/cdroms/micro2/content/education/wirth10.pdf)

    Parameters:
        polygon (Polygon):
            Input shapely polygon object.

    Returns:
        float:
            The length of the minor axis.
    """
    mrr = polygon.minimum_rotated_rectangle.exterior.coords
    return axis_len(mrr, "minor")


def major_axis_angle(polygon: Polygon) -> float:
    """Compute the major axis angle of a polygon.

    Note:
        The major axis is the (x,y) endpoints of the longest line that
        can be drawn through the object. Major axis angle is the angle of
        the major axis with respect to the x-axis.
        - [Wirth](http://www.cyto.purdue.edu/cdroms/micro2/content/education/wirth10.pdf)

    Parameters:
        polygon (Polygon):
            Input shapely polygon object.

    Returns:
        float:
            The angle of the major axis in degrees.
    """
    mrr = polygon.minimum_rotated_rectangle.exterior.coords
    return axis_angle(mrr, "major")


def minor_axis_angle(polygon: Polygon) -> float:
    """Compute the minor axis angle of a polygon.

    Note:
        The minor axis is the (x,y) endpoints of the longest line that
        can be drawn through the object whilst remaining perpendicular
        with the major-axis. Minor axis angle is the angle of the minor
        axis with respect to the x-axis.
        - [Wirth](http://www.cyto.purdue.edu/cdroms/micro2/content/education/wirth10.pdf)

    Parameters:
        polygon (Polygon):
            Input shapely polygon object.

    Returns:
        float:
            The angle of the minor axis in **degrees**.
    """
    mrr = polygon.minimum_rotated_rectangle.exterior.coords
    return axis_angle(mrr, "minor")


def compactness(polygon: Polygon, **kwargs) -> float:
    """Compute the compactness of a polygon.

    Note:
        Compactness is defined as the ratio of the area of an object
        to the area of a circle with the same perimeter. A circle is the
        most compact shape. Objects that are elliptical or have complicated,
        irregular (not smooth) boundaries have larger compactness.
        - [Wirth](http://www.cyto.purdue.edu/cdroms/micro2/content/education/wirth10.pdf)

    **Compactness:**
    $$
    \\frac{4\\pi A_{poly}}{P_{poly}^2}
    $$

    where $A_{poly}$ is the area of the polygon and $P_{poly}$ is the perimeter of
    the polygon.

    Parameters:
        polygon (Polygon):
            Input shapely polygon object.

    Returns:
        float:
            The compactness value of a polygon between 0-1.
    """
    perimeter = polygon.length
    area = polygon.area

    compactness = (4 * np.pi * area) / perimeter**2

    return compactness


def circularity(polygon: Polygon, **kwargs) -> float:
    """Compute the circularity of a polygon.

    Note:
        Circularity (sometimes roundness) is the ratio of the area of
        an object to the area of a circle with the same convex perimeter.
        Circularity equals 1 for a circular object and less than 1 for
        non-circular objects. Note that circularity is insensitive to
        irregular boundaries.
        - [Wirth](http://www.cyto.purdue.edu/cdroms/micro2/content/education/wirth10.pdf)

    **Circularity:**
    $$
    \\frac{4 \\times \\pi A_{poly}}{P_{convex}^2}
    $$

    where $A_{poly}$ is the area of the polygon and $P_{convex}$ is the perimeter of
    the convex hull.

    Parameters:
        polygon (Polygon):
            Input shapely polygon object.

    Returns:
        float:
            The circularity value of a polygon between 0-1.
    """
    convex_perimeter = polygon.convex_hull.length
    area = polygon.area

    circularity = (4 * np.pi * area) / convex_perimeter**2

    return circularity


def convexity(polygon: Polygon, **kwargs) -> float:
    """Compute the convexity of a polygon.

    Note:
        Convexity is the relative amount that an object differs from a
        convex object. Convexity is defined by computing the ratio of
        the perimeter of an object's convex hull to the perimeter of
        the object itself. This will take the value of 1 for a convex
        object, and will be less than 1 if the object is not convex, such
        as one having an irregular boundary.
        - [Wirth](http://www.cyto.purdue.edu/cdroms/micro2/content/education/wirth10.pdf)

    **Convexity:**
    $$
    \\frac{P_{convex}}{P_{poly}}
    $$

    where $P_{convex}$ is the perimeter of the convex hull and $P_{poly}$ is the
    perimeter of the polygon.

    Parameters:
        polygon (Polygon):
            Input shapely polygon object.

    Returns:
        float:
            The convexity value of a polygon between 0-1.
    """
    convex_perimeter = polygon.convex_hull.length
    perimeter = polygon.length

    convexity = convex_perimeter / perimeter

    return convexity


def solidity(polygon: Polygon, **kwargs) -> float:
    """Compute the solidity of a polygon.

    Note:
        Solidity measures the density of an object. It is defined as the
        ratio of the area of an object to the area of a convex hull of the
        object. A value of 1 signifies a solid object, and a value less than
        1 will signify an object having an irregular boundary, or containing
        holes.
        - [Wirth](http://www.cyto.purdue.edu/cdroms/micro2/content/education/wirth10.pdf)

    **Solidity:**
    $$
    \\frac{A_{poly}}{A_{convex}}
    $$

    where $A_{poly}$ is the area of the polygon and $A_{convex}$ is the area of the
    convex hull.

    Parameters:
        polygon (Polygon):
            Input shapely polygon object.

    Returns:
        float:
            The solidity value of a polygon between 0-1.
    """
    convex_area = polygon.convex_hull.area
    area = polygon.area

    return area / convex_area


def elongation(polygon: Polygon, **kwargs) -> float:
    """Compute the elongation of a polygon.

    Note:
        Elongation is the ratio between the length and width of the
        object bounding box. If the ratio is equal to 1, the object
        is roughly square or circularly shaped. As the ratio decreases
        from 1, the object becomes more elongated.
        - [Wirth](http://www.cyto.purdue.edu/cdroms/micro2/content/education/wirth10.pdf)

    **Elongation:**
    $$
    \\frac{\\text{bbox width}}{\\text{bbox height}}
    $$

    Parameters:
        polygon (Polygon):
            Input shapely polygon object.

    Returns:
        float:
            The elongation value of a polygon between 0-1.
    """
    minx, miny, maxx, maxy = polygon.bounds

    width = maxx - minx
    height = maxy - miny

    if width <= height:
        elongation = width / height
    else:
        elongation = height / width

    return elongation


def eccentricity(polygon: Polygon, **kwargs) -> float:
    """Compute the eccentricity of a polygon.

    Note:
        Eccentricity (sometimes ellipticity) measures how far the object is
        from an ellipse. It is defined as the ratio of the length of the minor
        axis to the length of the major axis of an object. The closer the
        object is to an ellipse, the closer the eccentricity is to 1
        - [Wirth](http://www.cyto.purdue.edu/cdroms/micro2/content/education/wirth10.pdf)

    **Eccentricity:**
    $$
    \\sqrt{1 - \\frac{\\text{minor axis}^2}{\\text{major axis}^2}}
    $$

    Parameters:
        polygon (Polygon):
            Input shapely polygon object.

    Returns:
        float:
            The eccentricity value of a polygon between 0-1.
    """
    mrr = polygon.minimum_rotated_rectangle.exterior.coords
    major_ax, minor_ax = axis_len(mrr)
    eccentricity = np.sqrt(1 - (minor_ax**2 / major_ax**2))
    return eccentricity


def fractal_dimension(polygon: Polygon, **kwargs) -> float:
    """Compute the fractal dimension of a polygon.

    Note:
        The fractal dimension is the rate at which the perimeter of an
        object increases as the measurement scale is reduced. The fractal
        dimension produces a single numeric value that summarizes the
        irregularity of "roughness" of the feature boundary.
        - [Wirth](http://www.cyto.purdue.edu/cdroms/micro2/content/education/wirth10.pdf)


    **Fractal dimension:**
    $$
    2 \\times \\frac{\\log(\\frac{P_{poly}}{4})}{\\log(A_{poly})}
    $$

    where $P_{poly}$ is the perimeter of the polygon and $A_{poly}$ is the area of the
    polygon.

    Parameters:
        polygon (Polygon):
            Input shapely polygon object.

    Returns:
        float:
            The fractal dimension value of a polygon.
    """
    perimeter = polygon.length
    area = polygon.area

    return (2 * np.log(perimeter / 4)) / np.log(area)


def sphericity(polygon: Polygon, **kwargs) -> float:
    """Compute the sphericity of a polygon.

    Note:
        Sphericity measures how close an object is to the shape of a “sphere”.
        It is defined as the ratio of the radius of the minimum inscribing circle
        to the radius of the minimum bounding circle.

    **Sphericity:**
    $$
    \\frac{\\text{MIR}}{\\text{MBR}}
    $$

    where $\\text{MIR}$ is the radius of the minimum inscribing circle radius and
    $\\text{MBR}$ is the radius of the minimum bounding radius.

    Parameters:
        polygon (Polygon):
            Input shapely polygon object.

    Returns:
        float:
            The sphericity value of a polygon.
    """
    _, _, ri = inscribing_circle(polygon)
    rc = shapely.minimum_bounding_radius(polygon)

    return ri / rc


def shape_index(polygon: Polygon, **kwargs) -> float:
    """Compute the shape index of a polygon.

    Note:
        Basically, the inverse of circularity.

    **Shape Index:**
    $$
    \\frac{\\sqrt{\\frac{A_{poly}}{\\pi}}}{\\text{MBR}}
    $$

    where $A_{poly}$ is the area of the polygon and $\\text{MBR}$ is the radius of the
    minimum bounding radius.

    Parameters:
        polygon (Polygon):
            Input shapely polygon object.

    Returns:
        float:
            The shape index value of a polygon.
    """
    r = shapely.minimum_bounding_radius(polygon)
    area = polygon.area

    return np.sqrt(area / np.pi) / r


def squareness(polygon: Polygon, **kwargs) -> float:
    """Compute the squareness of a polygon.

    Note:
        Squareness is a measure of how close an object is to a square.

    **Squareness:**
    $$
    \\left(\\frac{4*\\sqrt{A_{poly}}}{P_{poly}}\\right)^2
    $$

    where $A_{poly}$ is the area of the polygon and $P_{poly}$ is the perimeter of
    the polygon.

    Note:
        For irregular shapes, squareness is close to zero and for circular shapes close
        to 1.3. For squares, equals 1

    Parameters:
        polygon (Polygon):
            Input shapely polygon object.

    Returns:
        float:
            The squareness value of a polygon.
    """
    area = polygon.area
    perimeter = polygon.length

    return ((np.sqrt(area) * 4) / perimeter) ** 2


def rectangularity(polygon: Polygon, **kwargs) -> float:
    """Compute the rectangularity of a polygon.

    Note:
        Rectangularity is the ratio of the object to the area of the
        minimum bounding rectangle. Rectangularity has a value of 1
        for perfectly rectangular object.

    **Rectangularity:**
    $$
    \\frac{A_{poly}}{A_{MRR}}
    $$

    where $A_{poly}$ is the area of the polygon and $A_{MRR}$ is the area of the
    minimum rotated rectangle.

    Parameters:
        polygon (Polygon):
            Input shapely polygon object.

    Returns:
        float:
            The rectangularity value of a polygon between 0-1.
    """
    mrr = polygon.minimum_rotated_rectangle

    return polygon.area / mrr.area


def equivalent_rectangular_index(polygon: Polygon) -> float:
    """Compute the equivalent rectangular index.

    Note:
        Equivalent rectangluar index is the deviation of a polygon from
        an equivalent rectangle.

    **ERI:**
    $$
    \\frac{\\sqrt{A_{poly}}}{A_{MRR}}
    \\times
    \\frac{P_{MRR}}{P_{poly}}
    $$

    where $A_{poly}$ is the area of the polygon, $A_{MRR}$ is the area of the
    minimum rotated rectangle, $P_{MRR}$ is the perimeter of the minimum rotated
    rectangle and $P_{poly}$ is the perimeter of the polygon.

    Parameters:
        polygon (Polygon):
            Input shapely polygon object.

    Returns:
        float:
            The ERI value of a polygon between 0-1.
    """
    mrr = polygon.minimum_rotated_rectangle

    return np.sqrt(polygon.area / mrr.area) / (mrr.length / polygon.length)


SHAPE_LOOKUP = {
    "major_axis_len": major_axis_len,
    "minor_axis_len": minor_axis_len,
    "major_axis_angle": major_axis_angle,
    "minor_axis_angle": minor_axis_angle,
    "compactness": compactness,
    "circularity": circularity,
    "convexity": convexity,
    "solidity": solidity,
    "elongation": elongation,
    "eccentricity": eccentricity,
    "fractal_dimension": eccentricity,
    "sphericity": sphericity,
    "shape_index": shape_index,
    "rectangularity": rectangularity,
    "squareness": squareness,
    "equivalent_rectangular_index": equivalent_rectangular_index,
    "area": None,
}


def shape_metric(
    gdf: gpd.GeoDataFrame,
    metrics: Tuple[str, ...],
    parallel: bool = True,
    col_prefix: str = None,
    create_copy: bool = True,
) -> gpd.GeoDataFrame:
    """Compute a set of shape metrics for every row of the gdf.

    Parameters:
        gdf (gpd.GeoDataFrame):
            The input GeoDataFrame.
        metrics (Tuple[str, ...]):
            A Tuple/List of shape metrics.
        parallel (bool):
            Flag whether to use parallel apply operations when computing the diversities.
        col_prefix (str):
            Prefix for the new column names.
        create_copy (bool):
            Flag whether to create a copy of the input gdf or not.

    Note:
        Allowed shape metrics are:

        - `area`
        - `major_axis_len`
        - `minor_axis_len`
        - `major_axis_angle`
        - `minor_axis_angle`
        - `compactness`
        - `circularity`
        - `convexity`
        - `solidity`
        - `elongation`
        - `eccentricity`
        - `fractal_dimension`
        - `sphericity`
        - `shape_index`
        - `rectangularity`
        - `squareness`
        - `equivalent_rectangular_index`

    Raises:
        ValueError:
            If an illegal metric is given.

    Returns:
        gpd.GeoDataFrame:
            The input geodataframe with computed shape metric columns added.

    Examples:
        Compute the eccentricity and solidity for each polygon in gdf.
        >>> from cellseg_gsontools.geometry import shape_metric
        >>> shape_metric(gdf, metrics=["eccentricity", "solidity"], parallel=True)
    """
    if not isinstance(metrics, (list, tuple)):
        raise ValueError(f"`metrics` must be a list or tuple. Got: {type(metrics)}.")

    allowed = list(SHAPE_LOOKUP.keys())
    if not all(m in allowed for m in metrics):
        raise ValueError(
            f"Illegal metric in `metrics`. Got: {metrics}. Allowed metrics: {allowed}."
        )

    if create_copy:
        gdf = gdf.copy()

    if col_prefix is None:
        col_prefix = ""
    else:
        col_prefix += "_"

    met = list(metrics)
    if "area" in metrics:
        gdf[f"{col_prefix}area"] = gdf.area
        met.remove("area")

    for metric in met:
        gdf[metric] = gdf_apply(
            gdf,
            SHAPE_LOOKUP[metric],
            columns=["geometry"],
            parallel=parallel,
        )

    return gdf



from abc import ABC, abstractmethod
from typing import List, Union

import geopandas as gpd

__all__ = ["_SpatialBackend"]


class _SpatialBackend(ABC):
    @abstractmethod
    def convert_area_gdf(self):
        """Convert an area gdf to another gdf format."""
        raise NotImplementedError

    @abstractmethod
    def convert_cell_gdf(self):
        """Convert an area gdf to another gdf format."""
        raise NotImplementedError

    @abstractmethod
    def to_geopandas(self):
        """Convert a gdf to a geopandas dataframe."""
        raise NotImplementedError

    @abstractmethod
    def roi(self):
        """Get the region of interest."""
        raise NotImplementedError

    @abstractmethod
    def roi_cells(self):
        """Get the cells in the region of interest."""
        raise NotImplementedError

    @abstractmethod
    def interface(self):
        """Get the interface of two regions."""
        raise NotImplementedError

    def check_columns(
        self, area_gdf: gpd.GeoDataFrame, cell_gdf: gpd.GeoDataFrame
    ) -> gpd.GeoDataFrame:
        """Check if the input gdfs have the required columns.

        Parameters
        ----------
        area_gdf : gpd.GeoDataFrame
            A geodataframe containing the areas.
        cell_gdf : gpd.GeoDataFrame
            A geodataframe containing the cells.

        Returns
        -------
        gpd.GeoDataFrame:
            The gdf with the required columns.
        """
        if "class_name" not in area_gdf.columns:
            raise ValueError(
                "'class_name' column not found. `area_gdf` must contain a 'class_name' "
                f"column. Got: {list(area_gdf.columns)}"
            )

        if "class_name" not in cell_gdf.columns:
            raise ValueError(
                "'class_name' column not found. `cell_gdf` must contain a 'class_name' "
                f"column. Got: {list(cell_gdf.columns)}"
            )

    def filter_areas(
        self,
        area_gdf: gpd.GeoDataFrame,
        labels: Union[str, List[str]],
        min_area_size: float = None,
    ) -> gpd.GeoDataFrame:
        """Filter areas by tissue type and size.

        Parameters
        ----------
        area_gdf : gpd.GeoDataFrame
            A geodataframe containing the areas.
        labels : Union[str, List[str]]
            The labels to filter the areas by.
        min_area_size : float, optional
            The minimum area size to keep.

        Returns
        -------
        gpd.GeoDataFrame
            A geodataframe containing the filtered areas.
        """
        # get the areas that have type in labels
        if isinstance(labels, str):
            area_gdf = area_gdf[area_gdf["class_name"] == labels]
        else:
            if len(labels) == 1:
                area_gdf = area_gdf[area_gdf["class_name"] == labels[0]]
            else:
                area_gdf = area_gdf[area_gdf["class_name"].isin(labels)]

        # drop areas smaller than min_area_size
        if min_area_size is not None:
            area_gdf = area_gdf.loc[area_gdf.geometry.area >= min_area_size]

        return area_gdf



from pathlib import Path
from typing import Dict, Tuple, Union

from .save_utils import check_format, get_file_from_coords, get_xy_coords

__all__ = ["BaseGSONMerger"]


class BaseGSONMerger:
    def __init__(
        self, in_dir: Union[Path, str], tile_size: Tuple[int, int] = (1000, 1000)
    ) -> None:
        """Create a base class for geojson merger classes.

        Implements methods to access the adjascent tiles of a given tile.

        Parameters:
            in_dir (Union[Path, str]):
                Path to the directory containing the geojson files.
            tile_size (Tuple[int, int]):
                Height and width of the tile in pixels.
        """
        self.in_dir = Path(in_dir)
        self.tile_size = tile_size
        self.files = sorted(Path(in_dir).glob("*"))

        # Check that all the files are in the correct format
        if self.files:
            for f in self.files:
                check_format(f)
        else:
            raise ValueError(f"No files found in {in_dir}.")

    def _get_right_neighbor(self, fname: str) -> Path:
        x, y = get_xy_coords(fname)
        x += self.tile_size[0]
        return get_file_from_coords(self.files, x, y)

    def _get_left_neighbor(self, fname: str) -> Path:
        x, y = get_xy_coords(fname)
        x -= self.tile_size[0]
        return get_file_from_coords(self.files, x, y)

    def _get_bottom_neighbor(self, fname: str) -> Path:
        x, y = get_xy_coords(fname)
        y += self.tile_size[1]
        return get_file_from_coords(self.files, x, y)

    def _get_bottom_right_neighbor(self, fname: str) -> Path:
        x, y = get_xy_coords(fname)
        x += self.tile_size[0]
        y += self.tile_size[1]
        return get_file_from_coords(x, y)

    def _get_bottom_left_neighbor(self, fname: str) -> Path:
        x, y = get_xy_coords(fname)
        x -= self.tile_size[0]
        y += self.tile_size[1]
        return get_file_from_coords(self.files, x, y)

    def _get_top_neighbor(self, fname: str) -> Path:
        x, y = get_xy_coords(fname)
        y -= self.tile_size[1]
        return get_file_from_coords(self.files, x, y)

    def _get_top_right_neighbor(self, fname: str) -> Path:
        x, y = get_xy_coords(fname)
        x += self.tile_size[0]
        y -= self.tile_size[1]
        return get_file_from_coords(self.files, x, y)

    def _get_top_left_neighbor(self, fname: str) -> Path:
        x, y = get_xy_coords(fname)
        x -= self.tile_size[0]
        y -= self.tile_size[1]
        return get_file_from_coords(self.files, x, y)

    def _get_adjascent_tiles(self, fname: str) -> Dict[str, Path]:
        adj = {}
        adj["left"] = self._get_left_neighbor(fname)
        adj["right"] = self._get_right_neighbor(fname)
        adj["bottom"] = self._get_bottom_neighbor(fname)
        adj["top"] = self._get_top_neighbor(fname)

        return adj



import geopandas as gpd

from cellseg_gsontools.spatial_context.ops import get_interface_zones, get_objs

from ._base_backend import _SpatialBackend

__all__ = ["_SpatialContextGP"]


class _SpatialContextGP(_SpatialBackend):
    def __init__(self) -> None:
        """Create a geopandas spatial context class."""
        self.backend_name = "geopandas"

    def roi(self, context_area: gpd.GeoDataFrame, ix: int) -> gpd.GeoDataFrame:
        """Get a roi area of index `ix`.

        Parameters
        ----------
        context_area : gpd.GeoDataFrame
            The context area gdf.
        ix : int
            The index of the roi area. I.e., the ith roi area.

        Returns
        -------
            gpd.GeoDataFrame:
                The ith roi area.
        """
        row: gpd.GeoSeries = context_area.loc[ix]
        roi_area = gpd.GeoDataFrame([row], crs=context_area.crs)

        return roi_area

    def roi_cells(
        self,
        roi_area: gpd.GeoDataFrame,
        cell_gdf: gpd.GeoDataFrame,
        predicate: str = "intersects",
        silence_warnings: bool = True,
        **kwargs,
    ) -> gpd.GeoDataFrame:
        """Get the cells within the roi area.

        Parameters
        ----------
        roi_area : gpd.GeoDataFrame
            The roi area.
        cell_gdf : gpd.GeoDataFrame
            The cell gdf.
        predicate : str, default="intersects"
            The predicate to use for the spatial join. See `geopandas.tools.sjoin`
        silence_warnings : bool, default=True
            Flag, whether to silence all the geopandas related warnings.

        Returns
        -------
            gpd.GeoDataFrame:
                The cells within the roi area.
        """
        if roi_area is None or roi_area.empty:
            return

        objs_within: gpd.GeoDataFrame = get_objs(
            roi_area,
            cell_gdf,
            silence_warnings=silence_warnings,
            predicate=predicate,
        )

        if objs_within is None or objs_within.empty:
            return

        return objs_within.set_geometry("geometry")

    def interface(
        self,
        top_roi_area: gpd.GeoDataFrame,
        bottom_gdf: gpd.GeoDataFrame,
        buffer_dist: int,
    ) -> gpd.GeoDataFrame:
        """Get an interface area of index `ix`.

        Parameters
        ----------
            top_roi_area : gpd.GeoDataFrame
                The top roi area.
            bottom_gdf : gpd.GeoDataFrame
                The bottom gdf on top of which top_roi_area is buffered.
        """
        # Get the intersection of the roi and the area of type `label2`
        iface = get_interface_zones(
            buffer_area=top_roi_area,
            areas=bottom_gdf,
            buffer_dist=buffer_dist,
        )

        # If there are many interfaces, dissolve them into one
        if len(iface) > 1:
            iface = iface.dissolve().set_geometry("geometry")

        return iface

    def to_geopandas(self, gdf) -> gpd.GeoDataFrame:
        """Convenience identity function to match the backend API."""
        return gdf

    def convert_area_gdf(self, gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        """Convenience identity function to match the backend API.."""
        return gdf

    def convert_cell_gdf(self, gdf: gpd.GeoDataFrame, **kwargs) -> gpd.GeoDataFrame:
        """Convenience identity function to match the backend API.."""
        return gdf



from .shape_metrics import (
    SHAPE_LOOKUP,
    circularity,
    compactness,
    convexity,
    eccentricity,
    elongation,
    equivalent_rectangular_index,
    fractal_dimension,
    major_axis_angle,
    major_axis_len,
    minor_axis_angle,
    minor_axis_len,
    rectangularity,
    shape_index,
    shape_metric,
    solidity,
    sphericity,
    squareness,
)

__all__ = [
    "alpha_shape",
    "major_axis_len",
    "minor_axis_len",
    "major_axis_angle",
    "minor_axis_angle",
    "compactness",
    "circularity",
    "convexity",
    "solidity",
    "elongation",
    "eccentricity",
    "fractal_dimension",
    "sphericity",
    "shape_index",
    "rectangularity",
    "squareness",
    "equivalent_rectangular_index",
    "shape_metric",
    "SHAPE_LOOKUP",
]



from functools import partial
from typing import Any, Dict, Tuple, Union

import geopandas as gpd
import matplotlib.pyplot as plt
import pandas as pd
import psutil
from libpysal.weights import W, w_subset, w_union
from tqdm import tqdm

from cellseg_gsontools.apply import gdf_apply
from cellseg_gsontools.graphs import fit_graph
from cellseg_gsontools.grid import fit_spatial_grid
from cellseg_gsontools.links import weights2gdf
from cellseg_gsontools.plotting import plot_all
from cellseg_gsontools.spatial_context.backend import (
    _SpatialContextDGP,
    _SpatialContextGP,
    _SpatialContextSP,
)
from cellseg_gsontools.utils import set_uid

__all__ = ["WithinContext"]


class WithinContext:
    """Handle & extract cells from `cell_gdf` within the ROIs of `area_gdf`.

    Note:
        `area_gdf` and `cell_gdf` have to contain a column named 'class_name'

    Parameters:
        area_gdf (gpd.GeoDataFrame):
            A geo dataframe that contains large tissue area polygons enclosing
            the smaller cellular objects in `cell_gdf`.
        cell_gdf (gpd.GeoDataFrame):
            A geo dataframe that contains small cellular objects that are
            enclosed by larger tissue areas in `area_gdf`.
        labels (Union[Tuple[str, ...], str]):
            The class name(s) of the areas of interest. The objects within
            these areas are extracted. E.g. "cancer" or "stroma".
        min_area_size (float or str, optional):
            The minimum area of the objects that are kept. All the objects in
            the `area_gdf` that are larger are kept than `min_area_size`. If
            None, all the areas are kept. Defaults to None.
        graph_type (str):
            The type of the graph to be fitted to the cells inside interfaces.
            One of: "delaunay", "distband", "relative_nhood", "knn".
        dist_thresh (float):
            Distance threshold for the length of the network links.
        grid_type (str):
            The type of the grid to be fitted on the roi areas. One of:
            "square", "hex".
        patch_size (Tuple[int, int]):
            The size of the grid patches to be fitted on the context. This is
            used when `grid_type='square'`.
        stride (Tuple[int, int]):
            The stride of the sliding window for grid patching. This is used
            when `grid_type='square'`.
        pad (int):
            The padding to add to the bounding box on the grid. This is used
            when `grid_type='square'`.
        resolution (int):
            The resolution of the h3 hex grid. This is used when
            `grid_type='hex'`.
        predicate (str):
            The predicate to use for the spatial join when extracting the ROI
            cells. See `geopandas.tools.sjoin`
        silence_warnings (bool):
            Flag, whether to silence all the warnings.
        parallel (bool):
            Flag, whether to parallelize the context fitting. If
            `backend == "geopandas"`, the parallelization is implemented with
            `pandarallel` package. If `backend == "spatialpandas"`, or
            `backend == "dask-geopandas"` the parallelization is implemented
             with Dask library.
        num_processes (int):
            The number of processes to use when parallel=True. If -1, this
            will use all the available cores.
        backend (str):
            The backend to use for the spatial context. One of "geopandas",
            "spatialpandas" "dask-geopandas". "spatialpandas" or
            "dask-geopandas" is recommended for gdfs that may contain huge
            polygons.

    Attributes:
        context (Dict[int, Dict[str, Union[gpd.GeoDataFrame, libpysal.weights.W]]]):
            A nested dict that contains dicts for each of the distinct ROIs
            of type `label`. The keys of the outer dict are the indices of
            these areas. The inner dicts contain the keys:

            - `roi_area`- `gpd.GeoDataFrame`: the roi area. Roi area is the
                    tissue area of type `label`.
            - `roi_cells` - `gpd.GeoDataFrame`: the cells that are contained
                    inside the `roi_area`.
            - `roi_network` - `libpysal.weights.W`: spatial weights network of
                    the cells inside the `roi_area`. This can be used to extract
                    graph features inside the roi_.
            - `roi_grid` - `gpd.GeoDataFrame`: of the grid fitted on the `roi_area`.
                    This can be used to extract grid features inside the `roi_area`.

    Raises:
        ValueError:
            if `area_gdf` or `cell_gdf` don't contain 'class_name' column.
    """

    def __init__(
        self,
        area_gdf: gpd.GeoDataFrame,
        cell_gdf: gpd.GeoDataFrame,
        labels: Union[Tuple[str, ...], str],
        min_area_size: Union[float, str] = None,
        graph_type: str = "distband",
        dist_thresh: float = 100.0,
        grid_type: str = "square",
        patch_size: Tuple[int, int] = (256, 256),
        stride: Tuple[int, int] = (256, 256),
        pad: int = None,
        resolution: int = 9,
        predicate: str = "intersects",
        silence_warnings: bool = True,
        parallel: bool = False,
        num_processes: int = -1,
        backend: str = "geopandas",
    ) -> None:
        self.backend_name = backend
        if backend == "spatialpandas":
            self.backend = _SpatialContextSP()
        elif backend == "geopandas":
            self.backend = _SpatialContextGP()
        elif backend == "dask-geopandas":
            self.backend = _SpatialContextDGP()
        else:
            raise ValueError(
                f"Unknown backend: {backend}. "
                "Allowed: 'spatialpandas', 'geopandas', 'dask-geopandas'"
            )

        # check if the 'class_name' column is present
        self.backend.check_columns(area_gdf, cell_gdf)

        # set up the attributes
        self.min_area_size = min_area_size
        self.dist_thresh = dist_thresh
        self.graph_type = graph_type
        self.patch_size = patch_size
        self.stride = stride
        self.pad = pad
        self.silence_warnings = silence_warnings
        self.labels = labels
        self.predicate = predicate
        self.parallel = parallel
        self.num_processes = num_processes
        self.grid_type = grid_type
        self.resolution = resolution

        # set to geocentric cartesian crs. (unit is metre by default)
        # helps to avoid warning flooding
        self.cell_gdf = set_uid(cell_gdf, id_col="global_id")
        self.cell_gdf.set_crs(epsg=4328, inplace=True, allow_override=True)

        # cache the full area gdf for plotting
        self.area_gdf = area_gdf
        self.area_gdf.set_crs(epsg=4328, inplace=True, allow_override=True)

        # filter small areas and tissue types of interest for the tissue gdf
        self.context_area = self.backend.filter_areas(
            self.area_gdf, labels, min_area_size
        )
        self.context_area = set_uid(self.context_area, id_col="global_id")

        # set up cpu count
        if parallel:
            self.cpus = (
                psutil.cpu_count(logical=False)
                if self.num_processes == -1 or self.num_processes is None
                else self.num_processes
            )
        else:
            self.cpus = 1

        # convert the gdfs to the backend format
        self.context_area = self.backend.convert_area_gdf(self.context_area)

        self.cell_gdf = self.backend.convert_cell_gdf(
            self.cell_gdf, parallel=parallel, n_partitions=self.cpus
        )

    def __getattr__(self, name):
        """Get attribute."""
        return self.backend.__getattribute__(name)

    def fit(
        self,
        verbose: bool = True,
        fit_graph: bool = True,
        fit_grid: bool = True,
    ) -> None:
        """Fit the context.

        This sets the `self.context` attribute.

        Parameters:
            verbose (bool):
                Flag, whether to use tqdm pbar when creating the interfaces.
            fit_graph (bool):
                Flag, whether to fit the spatial weights networks for the
                context.
            fit_grid (bool):
                Flag, whether to fit the a grid on the contextes.

        Examples:
            Define a within context and plot the cells inside a specific ROI.

            >>> from cellseg_gsontools.spatial_context import WithinContext
            >>> area_gdf = read_gdf("area.json")
            >>> cell_gdf = read_gdf("cells.json")
            >>> within_context = WithinContext(
            ...     area_gdf=area_gdf,
            ...     cell_gdf=cell_gdf,
            ...     labels=["area_cin"],
            ...     silence_warnings=True,
            ...     min_area_size=100000.0,
            ... )
            >>> within_context.fit()
            >>> within_context.plot("roi_area", show_legends=True)
            <AxesSubplot: >
        """
        get_context_func = partial(
            WithinContext._get_context,
            backend=self.backend,
            context_area=self.context_area,
            cell_gdf=self.cell_gdf,
            fit_network=fit_graph,
            fit_grid=fit_grid,
            predicate=self.predicate,
            silence_warnings=self.silence_warnings,
            graph_type=self.graph_type,
            grid_type=self.grid_type,
            resolution=self.resolution,
            dist_thresh=self.dist_thresh,
            patch_size=self.patch_size,
            stride=self.stride,
            pad=self.pad,
            parallel=self.parallel,
            num_processes=self.cpus,
        )

        if self.backend_name == "geopandas" and self.parallel:
            # run in parallel
            context_dict = gdf_apply(
                self.context_area,
                func=get_context_func,
                columns=["global_id"],
                parallel=True,
                pbar=verbose,
                num_processes=self.cpus,
            ).to_dict()
        else:
            context_dict = {}
            pbar = (
                tqdm(self.context_area.index, total=self.context_area.shape[0])
                if verbose
                else self.context_area.index
            )

            for ix in pbar:
                if verbose:
                    pbar.set_description(f"Processing roi area: {ix}")

                if self.backend_name == "dask-geopandas" and self.parallel:
                    get_context_func = partial(
                        get_context_func, cell_gdf_dgp=self.backend.cell_gdf_dgp
                    )

                context_dict[ix] = get_context_func(ix=ix)

        self.context = context_dict

    @staticmethod
    def _get_context(
        ix: int,
        backend,
        context_area: gpd.GeoDataFrame,
        cell_gdf: gpd.GeoDataFrame,
        fit_network: bool = True,
        fit_grid: bool = True,
        predicate: str = "intersects",
        silence_warnings: bool = True,
        graph_type: str = "distband",
        grid_type: str = "square",
        resolution: int = 9,
        dist_thresh: float = 75.0,
        patch_size: Tuple[int, int] = (256, 256),
        stride: Tuple[int, int] = (256, 256),
        pad: int = None,
        parallel: bool = False,
        num_processes: int = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """Get the context dict of the given index."""
        roi_area: gpd.GeoDataFrame = backend.roi(ix=ix, context_area=context_area)
        roi_cells: gpd.GeoDataFrame = backend.roi_cells(
            roi_area=roi_area,
            cell_gdf=cell_gdf,
            predicate=predicate,
            silence_warnings=silence_warnings,
            parallel=parallel,
            num_processes=num_processes,
            **kwargs,
        )
        context_dict = {"roi_area": roi_area}
        context_dict["roi_cells"] = roi_cells

        if fit_network:
            roi_net: W = fit_graph(
                gdf=roi_cells,
                type=graph_type,
                id_col="global_id",
                thresh=dist_thresh,
                use_index=False,
            )
            context_dict["roi_network"] = roi_net

        if fit_grid:
            if grid_type == "hex":
                kwargs = {"resolution": resolution}
            else:
                kwargs = {
                    "patch_size": patch_size,
                    "stride": stride,
                    "pad": pad,
                    "predicate": predicate,
                }

            context_dict["roi_grid"] = fit_spatial_grid(
                gdf=roi_area, grid_type=grid_type, **kwargs
            )

        return context_dict

    def context2weights(self, key: str) -> W:
        """Merge the networks of type `key` into one spatial weights obj.

        Parameters:
            key (str):
                The key of the context dictionary that contains the spatial
                weights to be merged. One of "roi_network"

        Returns:
            libpysal.weights.W:
                A spatial weights object containing all the distinct networks
                in the context.
        """
        allowed = ("roi_network",)
        if key not in allowed:
            raise ValueError(f"Illegal key. Got: {key}. Allowed: {allowed}")

        cxs = list(self.context.items())
        wout = W({0: [0]})
        for _, c in cxs:
            w = c[key]
            if isinstance(w, W):
                wout = w_union(wout, w, silence_warnings=True)

        # remove self loops
        wout = w_subset(
            wout,
            list(wout.neighbors.keys())[1:],
            silence_warnings=True,
        )

        return wout

    def context2gdf(self, key: str) -> gpd.GeoDataFrame:
        """Merge the GeoDataFrames of type `key` into one geodataframe.

        Note:
            Returns None if no data is found.

        Parameters:
            key (str):
                The key of the context dictionary that contains the data to be
                converted to gdf. One of "roi_area", "roi_cells", "roi_grid",

        Returns:
            gpd.GeoDataFrame:
                Geo dataframe containing all the objects
        """
        allowed = (
            "roi_area",
            "roi_cells",
            "roi_grid",
        )
        if key not in allowed:
            raise ValueError(f"Illegal key. Got: {key}. Allowed: {allowed}")

        con = []
        for i in self.context.keys():
            if self.context[i][key] is not None:
                if isinstance(self.context[i][key], tuple):
                    con.append(self.context[i][key][0])
                else:
                    con.append(self.context[i][key])

        if not con:
            return

        gdf = pd.concat(
            con,
            keys=[i for i in self.context.keys() if self.context[i][key] is not None],
        )
        gdf = gdf.explode(ignore_index=True)

        return (
            gdf.reset_index(level=0, names="label")
            .drop_duplicates("geometry")
            .set_geometry("geometry")
        )

    def plot(
        self,
        key: str,
        network_key: str = None,
        grid_key: str = None,
        show_legends: bool = True,
        color: str = None,
        figsize: Tuple[int, int] = (12, 12),
        edge_kws: Dict[str, Any] = None,
        **kwargs,
    ) -> plt.Axes:
        """Plot the context with areas, cells, and ROIs highlighted.

        Parameters:
            key (str):
                The key of the context dictionary that contains the data to be plotted.
                One of "roi_area",
            network_key (str):
                The key of the context dictionary that contains the spatial weights to
                be plotted. One of "roi_network"
            grid_key (str):
                The key of the context dictionary that contains the grid to be plotted.
                One of "roi_grid"
            show_legends (bool):
                Flag, whether to include legends for each in the plot.
            color (str):
                A color for the interfaces or rois, Ignored if `show_legends=True`.
            figsize (Tuple[int, int]):
                Size of the figure.
            **kwargs (Dict[str, Any])]):
                Extra keyword arguments passed to the `plot` method of the
                geodataframes.

        Returns:
            AxesSubplot

        Examples:
            Plot the context with stromal areas highlighted.

            >>> from cellseg_gsontools.spatial_context import WithinContext
            >>> cells = read_gdf("cells.feather")
            >>> areas = read_gdf("areas.feather")
            >>> stroma = WithinContext(
            ...     cell_gdf=cells,
            ...     area_gdf=areas,
            ...     labels="stroma",
            ... )
            >>> stroma.fit(verbose=False)
            >>> stroma.plot("roi_area", show_legends=True)
            <AxesSubplot: >
        """
        allowed = "roi_area"
        if key != allowed:
            raise ValueError(f"Illegal key. Got: {key}. Allowed: {allowed}")

        context_gdf = self.context2gdf(key)

        grid_gdf = None
        if grid_key is not None:
            grid_gdf = self.context2gdf(grid_key)

        network_gdf = None
        if network_key is not None:
            edge_kws = edge_kws or {}
            w = self.context2weights(network_key)
            network_gdf = weights2gdf(self.cell_gdf, w)

        return plot_all(
            cell_gdf=self.cell_gdf.set_geometry("geometry"),
            area_gdf=self.area_gdf.set_geometry("geometry"),
            context_gdf=context_gdf,
            grid_gdf=grid_gdf,
            network_gdf=network_gdf,
            show_legends=show_legends,
            color=color,
            figsize=figsize,
            edge_kws=edge_kws,
            **kwargs,
        )



from typing import Tuple

import cv2
import numpy as np
from shapely.geometry import Polygon
from skimage.draw import polygon2mask
from skimage.morphology import erosion

__all__ = ["inscribing_circle", "circumscribing_circle"]


def inscribing_circle(polygon: Polygon) -> Tuple[float, float, float]:
    """Get the inscribing circle of a polygon.

    NOTE: sometimes for small polygons the center coords can be shifted
    but that does not affect the radius.

    Parameters
    ----------
        polygon : Polygon
            A shapely polygon object.

    Returns
    -------
        Tuple[float, float, float]:
            The centroid x, y- coords and the radius of the circle.
    """
    # get convex hull
    coords = polygon.exterior.coords

    # get hull bounds
    minx, miny, maxx, maxy = polygon.bounds
    w = maxx - minx
    h = maxy - miny
    shape = (int(h) + 1, int(w) + 1)

    # shift to origin for skimage
    points = np.array(coords)  # xy
    points[:, 0] -= minx
    points[:, 1] -= miny

    # convert poly to mask
    points = np.flip(points, 1)  # flip to yx for skimage
    m = np.pad(polygon2mask(shape, points), 1)  # add padding for dist transform
    m = erosion(m)  # erode

    # distance transform
    dist_map = cv2.distanceTransform(
        m.astype("uint8"), cv2.DIST_L2, cv2.DIST_MASK_PRECISE
    )
    dist_map = dist_map[1:-1, 1:-1]  # undo padding

    # find radius (max distance), and center (argmax)
    _, r, _, center = cv2.minMaxLoc(dist_map)

    # shift back to original coords
    x = center[0] + minx
    y = center[1] + miny

    return x, y, r


def circumscribing_circle(polygon: Polygon) -> Tuple[float, float, float]:
    """Get the circumscribing circle of a a polygon.

    Parameters
    ----------
        polygon : Polygon
            A shapely polygon object.

    Returns
    -------
        Tuple[float, float, float]:
            The centroid x, y- coords and the radius of the circle.
    """
    coords = polygon.exterior.coords
    center, r = cv2.minEnclosingCircle(np.array(coords).astype("int"))

    x = center[0]
    y = center[1]

    return x, y, r



from typing import List, Tuple

import geopandas as gpd
import pandas as pd
from shapely.geometry import Polygon
from shapely.ops import unary_union

__all__ = ["validate_and_simplify", "get_overlapping_objects", "merge_overlaps"]


def validate_and_simplify(
    poly: Polygon, buffer: float = 1.0, simplify: bool = False, level: float = 0.3
) -> Polygon:
    """Validate and simplify a shapely polygon.

    Parameters:
        poly (Polygon):
            The polygon to validate and simplify.
        buffer (float):
            The size of the buffer.
        simplify (bool):
            Whether to simplify the polygon or not.
        level (float):
            The simplification level.

    Returns:
        Polygon: The validated and simplified polygon.
    """
    # workaround to handle invalid polygons
    if not poly.is_valid:
        poly = poly.buffer(0)

    # do some simplifying
    if simplify:
        poly = (
            poly.buffer(10.0, join_style=1)
            .buffer(-10.0 + buffer, join_style=1)
            .simplify(level)
        )

    return poly


def get_overlapping_objects(gdf: gpd.GeoDataFrame) -> List[Tuple[int, int]]:
    """Get index pairs of overlapping objects in a gdf.

    Parameters:
        gdf (gpd.GeoDataFrame):
            The gdf to check for overlapping objects.

    Returns:
        (List[Tuple[int, int]]):
            A list of index pairs of overlapping objects.

    Examples:
        >>> import geopandas as gpd
        >>> from cellseg_gsontools.utils import read_gdf
        >>> gdf = read_gdf("/path/to/cells.json")
        >>> get_overlapping_objects(gdf)
        [(9, 34), (9, 7), (38, 39), (9, 34, 7), (32, 29, 31)]
    """
    overlap_inds = []
    for i in range(len(gdf)):
        overlaps = gdf.overlaps(gdf.geometry.loc[i])
        inds = set(i for i in overlaps[overlaps].index.tolist())
        inds.add(i)
        overlap_inds.append(inds)

    overlap_inds = list(set(frozenset(item) for item in overlap_inds if len(item) > 1))
    overlap_inds = [tuple(item) for item in overlap_inds]

    return overlap_inds


def merge_overlaps(gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """Merge overlapping objects in a gdf.

    Parameters:
        gdf (gpd.GeoDataFrame):
            The gdf containing overlapping objects.

    Returns:
        (gpd.GeoDataFrame):
            A gdf with merged overlapping objects.

    Examples:
        >>> import geopandas as gpd
        >>> from cellseg_gsontools.utils import read_gdf
        >>> gdf = read_gdf("/path/to/cells.json")
        >>> merge_overlaps(gdf)
    """
    # get overlapping objects
    gdf = gdf.copy()
    overlaps = get_overlapping_objects(gdf)

    # merge the overlapping objects
    new_polys = []
    new_classes = []
    for o in overlaps:
        # merge the overlapping objects and majority vote for the class
        # skip if the objects have already been merged
        if all([i in gdf.index for i in o]):
            data = gdf.loc[list(o)]
            new_polys.append(unary_union(data.geometry.tolist()))
            new_classes.append(data.class_name.value_counts().idxmax())

        # drop the objects from the orig gdf
        gdf = gdf.drop([i for i in o if i in gdf.index])

    overlap_gdf = gpd.GeoDataFrame({"geometry": new_polys, "class_name": new_classes})
    return pd.concat([gdf, overlap_gdf])



from functools import partial
from typing import Any, Callable, Tuple, Union

import geopandas as gpd
import pandas as pd
import shapely
from shapely.geometry import Polygon, box, mapping

from cellseg_gsontools.apply import gdf_apply
from cellseg_gsontools.utils import get_holes, lonlat_to_xy, xy_to_lonlat

try:
    import h3

    _has_h3 = True
except ImportError:
    _has_h3 = False

__all__ = [
    "bounding_box",
    "get_grid",
    "grid_overlay",
    "hexgrid_overlay",
    "fit_spatial_grid",
]


def fit_spatial_grid(
    gdf: gpd.GeoDataFrame, grid_type: str = "square", **kwargs
) -> gpd.GeoDataFrame:
    """Quick wrapper to fit either a hex or square grid to a `geopandas.GeoDataFrame`.

    Note:
        - Hexagonal grid requires the `h3` package to be installed.
        - Hexagonal grid only works for a gdf containing one single polygon.

    Parameters:
        gdf (gpd.GeoDataFrame):
            GeoDataFrame to fit grid to.
        grid_type (str):
            Type of grid to fit, by default "square".
        **kwargs (Dict[str, Any]):
            Keyword arguments to pass to grid fitting functions.

    Returns:
        gpd.GeoDataFrame:
            Fitted grid.

    Raises:
        ValueError: If grid_type is not one of "square" or "hex".
        ImportError: If grid_type is "hex" and the `h3` package is not installed.

    Examples:
        Fit a hexagonal grid to a gdf:
        >>> from cellseg_gsontools import read_gdf
        >>> from cellseg_gsontools.grid import fit_spatial_grid
        >>> # Read in the tissue areas
        >>> area_gdf = gpd.read_file("path/to/area.geojson")
        >>> # Fit the grid
        >>> hex_grid = fit_spatial_grid(area_gdf, grid_type="hex", resolution=9)
        >>> hex_grid
        gpd.GeoDataFrame

        Fit a square grid to a gdf:
        >>> from cellseg_gsontools import read_gdf
        >>> from cellseg_gsontools.grid import fit_spatial_grid
        >>> # Read in the tissue areas
        >>> area_gdf = gpd.read_file("path/to/area.geojson")
        >>> # Fit the grid
        >>> sq_grid = fit_spatial_grid(
        ...     area_gdf, grid_type="square", patch_size=(256, 256), stride=(256, 256)
        ... )
        >>> sq_grid
        gpd.GeoDataFrame
    """
    allowed = ["square", "hex"]
    if grid_type not in allowed:
        raise ValueError(f"grid_type must be one of {allowed}, got {grid_type}")

    if grid_type == "square":
        grid = grid_overlay(gdf, **kwargs)
    else:
        if not _has_h3:
            raise ImportError("h3 package not installed. Install with `pip install h3`")
        grid = hexgrid_overlay(gdf, **kwargs)

    return grid


def hexgrid_overlay(
    gdf: gpd.GeoDataFrame, resolution: int = 9, to_lonlat: bool = True
) -> gpd.GeoDataFrame:
    """Fit a `h3` hexagonal grid on top of a `geopandas.GeoDataFrame`.

    Parameters:
        gdf (gpd.GeoDataFrame):
            GeoDataFrame to fit grid to.
        resolution (int):
            H3 resolution, by default 9.
        to_lonlat (bool):
            Whether to convert to lonlat coordinates, by default True.

    Returns:
        gpd.GeoDataFrame:
            Fitted h3 hex grid.

    Examples:
        Fit a hexagonal grid to a gdf:
        >>> from cellseg_gsontools import read_gdf
        >>> from cellseg_gsontools.grid import hexgrid_overlay
        >>> # Read in the tissue areas
        >>> area_gdf = gpd.read_file("path/to/area.geojson")
        >>> # Fit the grid
        >>> hex_grid = hexgrid_overlay(area_gdf, resolution=9)
        >>> hex_grid
        gpd.GeoDataFrame
    """
    if gdf.empty or gdf is None:
        return

    orig_crs = gdf.crs

    poly = shapely.force_2d(gdf.unary_union)
    if isinstance(poly, Polygon):
        hexagons = poly2hexgrid(poly, resolution=resolution, to_lonlat=to_lonlat)
    else:
        output = []
        for geom in poly.geoms:
            hexes = poly2hexgrid(geom, resolution=resolution, to_lonlat=to_lonlat)
            output.append(hexes)
        hexagons = pd.concat(output)

    return hexagons.set_crs(orig_crs, inplace=True, allow_override=True)


def grid_overlay(
    gdf: gpd.GeoDataFrame,
    patch_size: Tuple[int, int] = (256, 256),
    stride: Tuple[int, int] = (256, 256),
    pad: int = 20,
    predicate: str = "intersects",
) -> gpd.GeoDataFrame:
    """Overlay a square grid to the given areas of a `geopandas.GeoDataFrame`.

    Note:
        Returns None if the gdf is empty.

    Parameters:
        gdf (gpd.GeoDataFrame):
            GeoDataFrame to fit the grid to. Uses the bounding box of the GeoDataFrame
            to fit the grid.
        patch_size (Tuple[int, int]):
            Patch size of the grid.
        stride (Tuple[int, int]):
            Stride of the sliding window in the grid.
        pad (int):
            Pad the bounding box with the given number of pixels, by default None.
        predicate (str):
            Predicate to use for the spatial join, by default "intersects".
            Allowed values are "intersects" and "within".

    Returns:
        gpd.GeoDataFrame:
            GeoDataFrame with the grid fitted to the given GeoDataFrame.

    Raises:
        ValueError: If predicate is not one of "intersects" or "within".

    Examples:
        Fit a square grid to a gdf:
        >>> from cellseg_gsontools import read_gdf
        >>> from cellseg_gsontools.grid import grid_overlay
        >>> # Read in the tissue areas
        >>> area_gdf = gpd.read_file("path/to/area.geojson")
        >>> # Fit the grid
        >>> sq_grid = grid_overlay(area_gdf, patch_size=(256, 256), stride=(256, 256))
        >>> sq_grid
        gpd.GeoDataFrame
    """
    if gdf.empty or gdf is None:
        return

    allowed = ["intersects", "within"]
    if predicate not in allowed:
        raise ValueError(f"predicate must be one of {allowed}. Got {predicate}")
    grid = get_grid(gdf, patch_size, stride, pad=pad)
    grid.set_crs(epsg=4328, inplace=True, allow_override=True)
    grid = grid.sjoin(gdf, predicate=predicate)

    return grid.drop_duplicates("geometry")


def poly2hexgrid(
    poly: Polygon, resolution: int = 9, to_lonlat: bool = True
) -> gpd.GeoDataFrame:
    """Convert a shapely Polygon to a h3 hexagon grid.

    Parameters:
        poly (Polygon):
            Polygon to convert.
        resolution (int, optional):
            H3 resolution, by default 9.
        to_lonlat (bool, optional):
            Whether to convert to lonlat coordinates, by default True.

    Returns:
        gpd.GeoDataFrame:
            GeoDataFrame of h3 hexagons.
    """
    x, y = poly.exterior.coords.xy
    if to_lonlat:
        x, y = xy_to_lonlat(x, y)
    holes = get_holes(poly, to_lonlat=to_lonlat)

    poly = Polygon(list(zip(x, y)), holes=holes)
    hexs = h3.polyfill(mapping(poly), resolution, geo_json_conformant=True)

    # to gdf
    hex_polys = gpd.GeoSeries(list(map(polygonise, hexs)), index=hexs, crs=4326)
    hex_polys = gpd.GeoDataFrame(geometry=hex_polys)

    return hex_polys


def polygonise(hex_id: str, to_cartesian: bool = True) -> Polygon:
    """Polygonise a h3 hexagon.

    Parameters:
        hex_id (str):
            H3 hexagon id.
        to_cartesian (bool, optional):
            Whether to convert to cartesian coordinates, by default True.

    Returns:
        Polygon:
            Polygonised h3 hexagon.
    """
    poly = Polygon(h3.h3_to_geo_boundary(hex_id, geo_json=True))

    if to_cartesian:
        lon, lat = poly.exterior.coords.xy
        x, y = lonlat_to_xy(lon, lat)
        poly = Polygon(list(zip(x, y)))

    return poly


def bounding_box(gdf: gpd.GeoDataFrame, pad: int = 0) -> gpd.GeoDataFrame:
    """Get the bounding box of a GeoDataFrame.

    Note:
        returns None if the gdf is empty.

    Parameters:
        gdf (gpd.GeoDataFrame):
            The input GeoDataFrame.
        pad (int):
            The padding to add to the bounding box.

    Returns:
        gpd.GeoDataFrame:
            The bounding box as a GeoDataFrame.
    """
    if gdf.empty or gdf is None:
        return

    xmin, ymin, xmax, ymax = gdf.total_bounds
    bbox = box(xmin - pad, ymin - pad, xmax + pad, ymax + pad)
    return gpd.GeoDataFrame({"geometry": bbox}, index=[0])


def _get_margins(
    first_endpoint: int, size: int, stride: int, pad: int = None
) -> Tuple[int, int]:
    """Get the number of slices needed for one direction and the overlap.

    Parameters:
        first_endpoint (int):
            The first coordinate of the patch.
        size (int):
            The size of the input.
        stride (int):
            The stride of the sliding window.
        pad (int):
            The padding to add to the patch

    Returns:
        Tuple[int, int]:
            The number of patches needed for one direction and the overlap.
    """
    pad = int(pad) if pad is not None else 20  # at least some padding needed
    size += pad

    n = 1
    mod = 0
    end = first_endpoint
    while True:
        n += 1
        end += stride

        if end > size:
            mod = end - size
            break
        elif end == size:
            break

    return n, mod + pad


def get_grid(
    gdf: gpd.GeoDataFrame,
    patch_size: Tuple[int, int] = (256, 256),
    stride: Tuple[int, int] = (256, 256),
    pad: int = 20,
) -> gpd.GeoDataFrame:
    """Get a grid of patches from a GeoDataFrame.

    Note:
        returns None if the gdf is empty.

    Parameters:
        gdf (gpd.GeoDataFrame):
            The input GeoDataFrame.
        patch_size (Tuple[int, int]):
            The size of the patch.
        stride (Tuple[int, int]):
            The stride of the sliding window.
        pad (int):
            The padding to add to the patch

    Returns:
        gpd.GeoDataFrame:
            The grid of patches.
    """
    if gdf.empty or gdf is None:
        return

    # add some default padding
    if pad is None:
        pad = 20

    bbox: gpd.GeoDataFrame = bounding_box(gdf, pad=pad)
    minx, miny, maxx, maxy = bbox.geometry.bounds.values[0]
    width = maxx - minx
    height = maxy - miny

    total_size = (height, width)

    y_end, x_end = patch_size
    nrows, _ = _get_margins(y_end, total_size[0], stride[0], pad=pad)
    ncols, _ = _get_margins(x_end, total_size[1], stride[1], pad=pad)

    grid_rects = []
    for row in range(nrows):
        for col in range(ncols):
            y_start = row * stride[0] + miny
            y_end = y_start + patch_size[0]
            x_start = col * stride[1] + minx
            x_end = x_start + patch_size[1]
            rect = Polygon(
                [(x_start, y_start), (x_end, y_start), (x_end, y_end), (x_start, y_end)]
            )
            grid_rects.append(rect)

    return gpd.GeoDataFrame({"geometry": grid_rects})


def get_rect_metric(
    rect, objs: gpd.GeoDataFrame, metric_func: Callable, predicate: str
) -> Any:
    """Get the metric of the given rectangle.

    Parameters:
        rect (Polygon):
            The rectangle to get the metric of.
        objs (gpd.GeoDataFrame):
            The objects to use for the metric.
        metric_func (Callable):
            The metric function to use.
        predicate (str):
            The predicate to use for the spatial join. Allowed values are "intersects"
            and "within".

    Returns:
        Any:
            The metric of the rectangle.
    """
    allowed = ["intersects", "within"]
    if predicate not in allowed:
        raise ValueError(f"predicate must be one of {allowed}. Got {predicate}.")

    sub_objs = objs.iloc[objs.geometry.sindex.query(rect, predicate=predicate)]

    return metric_func(sub_objs)


def grid_classify(
    grid: gpd.GeoDataFrame,
    objs: gpd.GeoDataFrame,
    metric_func: Callable,
    predicate: str,
    new_col_names: Union[Tuple[str, ...], str],
    parallel: bool = True,
    num_processes: int = -1,
    pbar: bool = False,
    **kwargs,
) -> gpd.GeoDataFrame:
    """Classify the grid based on objs inside the grid cells.

    Parameters:
        grid (gpd.GeoDataFrame):
            The grid of rectangles to classify.
        objs (gpd.GeoDataFrame):
            The objects to use for classification.
        metric_func (Callable):
            The metric/heuristic function to use for classification.
        predicate (str):
            The predicate to use for the spatial join. Allowed values are "intersects"
            and "within".
        new_col_names (Union[Tuple[str, ...], str]):
            The name of the new column(s) in the grid gdf.
        parallel (bool):
            Whether to use parallel processing.
        num_processes (int):
            The number of processes to use. If -1, uses all available cores.
            Ignored if parallel=False.
        pbar (bool):
            Whether to show a progress bar. Ignored if parallel=False.

    Returns:
        gpd.GeoDataFrame:
            The grid with the new columns added.

    Raises:
        ValueError: If predicate is not one of "intersects" or "within".

    Examples:
        Get the number of immune cells in each grid cell at the tumor stroma interface:
        >>> import geopandas as gpd
        >>> from shapely.geometry import Polygon
        >>> from functools import partial
        >>> from cellseg_gsontools import gdf_apply, read_gdf
        >>> from cellseg_gsontools.grid import grid_classify, grid_overlay
        >>> from cellseg_gsontools.context import InterfaceContext
        >>> # Define a heuristic function to get the number of immune cells
        >>> def get_immune_cell_cnt(gdf: gpd.GeoDataFrame, **kwargs) -> int:
        ...     try:
        ...         cnt = gdf.class_name.value_counts()["inflammatory"]
        ...     except KeyError:
        ...         cnt = 0
        ...     return int(cnt)
        >>> # Read in the tissue areas and cells
        >>> area_gdf = gpd.read_file("path/to/area.geojson")
        >>> cell_gdf = gpd.read_file("path/to/cell.geojson")
        >>> # Fit a tumor-stroma interface
        >>> tumor_stroma_iface = InterfaceContext(
        ...     area_gdf=area_gdf,
        ...     cell_gdf=cell_gdf,
        ...     top_labels="area_cin",
        ...     bottom_labels="areastroma",
        ...     buffer_dist=250,
        ...     graph_type="distband",
        ...     dist_thresh=75,
        ...     patch_size=(128, 128),
        ...     stride=(128, 128),
        ...     min_area_size=50000,
        ... )
        >>> tumor_stroma_iface.fit(verbose=False)
        >>> # Get the grid and the cells at the interface
        >>> iface_grid = grid_overlay(
        ...     tumor_stroma_iface.context2gdf("interface_area"),
        ...     patch_size=(128, 128),
        ...     stride=(128, 128),
        ... )
        >>> cells = tumor_stroma_iface.context2gdf("interface_cells")
        >>> # Classify the grid
        >>> iface_grid = grid_classify(
        >>>     grid=iface_grid,
        >>>     objs=cells,
        >>>     metric_func=get_immune_cnt,
        >>>     predicate="intersects",
        >>>     new_col_name="immune_cnt",
        >>>     parallel=True,
        >>>     pbar=True,
        >>>     num_processes=-1
        >>> )
        >>> iface_grid
        geometry  immune_cnt
        28  POLYGON ((20032.00000 54098.50000, 20160.00000... 15
        29  POLYGON ((20160.00000 54098.50000, 20288.00000... 3
    """
    allowed = ["intersects", "within"]
    if predicate not in allowed:
        raise ValueError(f"predicate must be one of {allowed}. Got {predicate}")

    if isinstance(new_col_names, str):
        new_col_names = [new_col_names]

    func = partial(
        get_rect_metric, objs=objs, predicate=predicate, metric_func=metric_func
    )
    grid.loc[:, list(new_col_names)] = gdf_apply(
        grid,
        func=func,
        parallel=parallel,
        pbar=pbar,
        num_processes=num_processes,
        columns=["geometry"],
    )

    return grid



from functools import partial
from typing import Any, Dict, Tuple, Union

import geopandas as gpd
import matplotlib.pyplot as plt
import pandas as pd
import psutil
from libpysal.weights import W, w_subset, w_union
from tqdm import tqdm

from cellseg_gsontools.apply import gdf_apply
from cellseg_gsontools.graphs import fit_graph, get_border_crosser_links
from cellseg_gsontools.grid import fit_spatial_grid
from cellseg_gsontools.links import weights2gdf
from cellseg_gsontools.plotting import plot_all
from cellseg_gsontools.spatial_context.backend import (
    _SpatialContextDGP,
    _SpatialContextGP,
    _SpatialContextSP,
)
from cellseg_gsontools.utils import set_uid

__all__ = ["InterfaceContext"]


class InterfaceContext:
    """Handle & extract interface regions from the `cell_gdf` and `area_gdf`.

    Interfaces are created by buffering areas of type `top_label` on top of the
    areas of type `bottom_labels` and taking the intersection of the buffered area
    and the `top_label` area. The result interface is a band-like area b/w
    `top_label` and `bottom_label` areas. The bredth of the interface is given by
    the `buffer_dist` param.

    Note:
        `area_gdf` and `cell_gdf` have to contain a column named 'class_name'

    Parameters:
        area_gdf (gpd.GeoDataFrame):
            A geo dataframe that contains large tissue area polygons enclosing
            the smaller cellular objects in `cell_gdf`.
        cell_gdf (gpd.GeoDataFrame):
            A geo dataframe that contains small cellular objects that are
            enclosed by larger tissue areas in `area_gdf`.
        top_labels (Union[Tuple[str, ...], str]):
            The class name(s) of the areas of interest. E.g. "tumor". These areas
            are buffered on top of the areas that have type in `bottom_labels`. For
            example, buffering the tumor area on top of the stroma will get the
            tumor-stroma interface.
        bottom_labels (Union[Tuple[str, ...], str]):
            The class name of the area on top of which the buffering is applied.
            Typically you want to buffer at least on top of the stromal area to get
            e.g. tumor-stroma interface. Other options are ofc possible.
        min_area_size (float or str, optional):
            The minimum area of the objects that are kept. All the objects in
            the `area_gdf` that are larger are kept than `min_area_size`. If
            None, all the areas are kept. Defaults to None.
        graph_type (str):
            The type of the graph to be fitted to the cells inside interfaces.
            One of: "delaunay", "distband", "relative_nhood", "knn".
        dist_thresh (float):
            Distance threshold for the length of the network links.
        grid_type (str):
            The type of the grid to be fitted on the roi areas. One of:
            "square", "hex".
        patch_size (Tuple[int, int]):
            The size of the grid patches to be fitted on the context. This is
            used when `grid_type='square'`.
        stride (Tuple[int, int]):
            The stride of the sliding window for grid patching. This is used
            when `grid_type='square'`.
        pad (int):
            The padding to add to the bounding box on the grid. This is used
            when `grid_type='square'`.
        resolution (int):
            The resolution of the h3 hex grid. This is used when
            `grid_type='hex'`.
        predicate (str):
            The predicate to use for the spatial join when extracting the ROI
            cells. See `geopandas.tools.sjoin`
        silence_warnings (bool):
            Flag, whether to silence all the warnings.
        parallel (bool):
            Flag, whether to parallelize the context fitting. If
            `backend == "geopandas"`, the parallelization is implemented with
            `pandarallel` package. If `backend == "spatialpandas"`, or
            `backend == "dask-geopandas"` the parallelization is implemented
            with Dask library.
        num_processes (int):
            The number of processes to use when parallel=True. If -1, this
            will use all the available cores.
        backend (str):
            The backend to use for the spatial context. One of "geopandas",
            "spatialpandas" "dask-geopandas". "spatialpandas" or
            "dask-geopandas" is recommended for gdfs that may contain huge
            polygons.

    Attributes:
        context (Dict[int, Dict[str, Union[gpd.GeoDataFrame, libpysal.weights.W]]]):
            A nested dict that contains dicts for each of the distinct ROIs
            of type `top_labels` and the interfaces b/w areas of type `bottom_labels`.
            The keys of the outer dict are the indices of these areas.
            The inner dicts contain the keys:

            - `roi_area`- `gpd.GeoDataFrame`: of the roi area. Roi area is the tissue
                    area(s) of type `top_labels` that is buffered on top of the area
                    of type `bottom_labels` to get the interface.
            - `roi_cells` - `gpd.GeoDataFrame`: of the cells that are contained
                    inside the `roi_area`.
            - `roi_network` - `libpysal.weights.W`: spatial weights network of
                    the cells inside the `roi_area`. This can be used to extract
                    graph features inside the `roi_area`.
            - `roi_grid` - `gpd.GeoDataFrame`: of the grid fitted on the `roi_area`.
                    This can be used to extract grid features inside the `roi_area`.
            - `interface_area` - `gpd.GeoDataFrame`:the interface area. Interface
                    area is the area that is the intersection of the buffered
                    `roi_area` (`top_labels`) and the area of type `bottom_labels`.
            - `interface_network` - `libpysal.weights.W`: spatial weights network of
                    the cells inside the `interface_area`.
            - `border_network` - `libpysal.weights.W`: spatial weights network of the
                    cells at the border of the roi and interface areas.
            - `full_network` - `libpysal.weights.W`: spatial weights network of the
                    cells inside the union of the roi and interface areas.

    Raises:
        ValueError: if `area_gdf` or `cell_gdf` don't contain 'class_name' column.
    """

    def __init__(
        self,
        area_gdf: gpd.GeoDataFrame,
        cell_gdf: gpd.GeoDataFrame,
        top_labels: Union[Tuple[str, ...], str],
        bottom_labels: Union[Tuple[str, ...], str],
        min_area_size: Union[float, str] = None,
        buffer_dist: int = 200,
        graph_type: str = "distband",
        dist_thresh: float = 50.0,
        grid_type: str = "square",
        patch_size: Tuple[int, int] = (256, 256),
        stride: Tuple[int, int] = (256, 256),
        pad: int = None,
        resolution: int = 9,
        predicate: str = "intersects",
        silence_warnings: bool = True,
        parallel: bool = False,
        num_processes: int = -1,
        backend: str = "geopandas",
    ) -> None:
        self.backend_name = backend
        if backend == "spatialpandas":
            self.backend = _SpatialContextSP()
        elif backend == "geopandas":
            self.backend = _SpatialContextGP()
        elif backend == "dask-geopandas":
            self.backend = _SpatialContextDGP()
        else:
            raise ValueError(
                f"Unknown backend: {backend}. "
                "Allowed: 'spatialpandas', 'geopandas', 'dask-geopandas'"
            )

        # check if the 'class_name' column is present
        self.backend.check_columns(area_gdf, cell_gdf)

        # set up the attributes
        self.buffer_dist = buffer_dist
        self.min_area_size = min_area_size
        self.dist_thresh = dist_thresh
        self.graph_type = graph_type
        self.patch_size = patch_size
        self.stride = stride
        self.pad = pad
        self.silence_warnings = silence_warnings
        self.top_labels = top_labels
        self.bottom_labels = bottom_labels
        self.predicate = predicate
        self.parallel = parallel
        self.num_processes = num_processes
        self.grid_type = grid_type
        self.resolution = resolution

        # set to geocentric cartesian crs. (unit is metre not degree as by default)
        # helps to avoid warning flooding
        self.cell_gdf = set_uid(cell_gdf, id_col="global_id")
        self.cell_gdf.set_crs(epsg=4328, inplace=True, allow_override=True)

        # cache the full area gdf for plotting
        self.area_gdf = area_gdf
        self.area_gdf.set_crs(epsg=4328, inplace=True, allow_override=True)

        # filter small areas and tissue types for the top and bottom labels
        self.context_area = self.backend.filter_areas(
            self.area_gdf, top_labels, min_area_size
        )
        self.context_area = set_uid(self.context_area, id_col="global_id")

        self.context_area2 = self.backend.filter_areas(
            self.area_gdf, bottom_labels, min_area_size
        )
        self.context_area2 = set_uid(self.context_area2, id_col="global_id")

        # set up cpu count
        if parallel:
            self.cpus = (
                psutil.cpu_count(logical=False)
                if self.num_processes == -1 or self.num_processes is None
                else self.num_processes
            )
        else:
            self.cpus = 1

        # convert the gdfs to the backend format
        self.context_area = self.backend.convert_area_gdf(self.context_area)
        self.context_area2 = self.backend.convert_area_gdf(self.context_area2)

        self.cell_gdf = self.backend.convert_cell_gdf(
            self.cell_gdf, parallel=parallel, n_partitions=self.cpus
        )

    def __getattr__(self, name):
        """Get attribute."""
        return self.backend.__getattribute__(name)

    def fit(
        self,
        verbose: bool = True,
        fit_graph: bool = True,
        fit_grid: bool = True,
    ) -> None:
        """Fit the interface context.

        This sets the `self.context` class attribute.

        Parameters:
            verbose (bool):
                Flag, whether to use tqdm pbar when creating the interfaces.
            fit_graph (bool):
                Flag, whether to fit the spatial weights networks for the
                context.
            fit_grid (bool):
                Flag, whether to fit the a grid on the contextes.

        Examples:
            Define an tumor-stroma interface context and plot the cells inside the
            interface area.

            >>> from cellseg_gsontools.backend import InterfaceContext
            >>> area_gdf = read_gdf("area.json")
            >>> cell_gdf = read_gdf("cells.json")
            >>> interface_context = InterfaceContext(
            ...     area_gdf=area_gdf,
            ...     cell_gdf=cell_gdf,
            ...     top_labels=["area_cin"],
            ...     bottom_labels=["area_stroma"],
            ...     buffer_dist=250.0,
            ...     graph_type="delaunay",
            ...     silence_warnings=True,
            ...     min_area_size=100000.0,
            ... )
            >>> interface_context.fit(parallel=False)
            >>> interface_context.plot("interface_area", show_legends=True)
            <AxesSubplot: >
        """
        get_context_func = partial(
            InterfaceContext._get_context,
            backend=self.backend,
            context_area=self.context_area,
            context_area2=self.context_area2,
            cell_gdf=self.cell_gdf,
            fit_network=fit_graph,
            fit_grid=fit_grid,
            grid_type=self.grid_type,
            resolution=self.resolution,
            predicate=self.predicate,
            buffer_dist=self.buffer_dist,
            silence_warnings=self.silence_warnings,
            graph_type=self.graph_type,
            dist_thresh=self.dist_thresh,
            patch_size=self.patch_size,
            stride=self.stride,
            pad=self.pad,
            parallel=self.parallel,
            num_processes=self.cpus,
        )

        if self.backend_name == "geopandas" and self.parallel:
            # run in parallel
            context_dict = gdf_apply(
                self.context_area,
                func=get_context_func,
                columns=["global_id"],
                parallel=True,
                pbar=verbose,
                num_processes=self.cpus,
            ).to_dict()
        else:
            context_dict = {}
            pbar = (
                tqdm(self.context_area.index, total=self.context_area.shape[0])
                if verbose
                else self.context_area.index
            )

            for ix in pbar:
                if verbose:
                    pbar.set_description(f"Processing roi area: {ix}")

                if self.backend_name == "dask-geopandas" and self.parallel:
                    get_context_func = partial(
                        get_context_func, cell_gdf_dgp=self.backend.cell_gdf_dgp
                    )

                context_dict[ix] = get_context_func(ix=ix)

        self.context = context_dict

    @staticmethod
    def _get_context(
        ix: int,
        backend,
        context_area: gpd.GeoDataFrame,
        context_area2: gpd.GeoDataFrame,
        cell_gdf: gpd.GeoDataFrame,
        buffer_dist: int = 200,
        fit_network: bool = True,
        fit_grid: bool = True,
        grid_type: str = "square",
        resolution: int = 9,
        predicate: str = "intersects",
        silence_warnings: bool = True,
        graph_type: str = "distband",
        dist_thresh: float = 75.0,
        patch_size: Tuple[int, int] = (256, 256),
        stride: Tuple[int, int] = (256, 256),
        pad: int = None,
        parallel: bool = False,
        num_processes: int = None,
        **kwargs,
    ) -> Dict[int, Any]:
        """Get the context dict of the given index."""
        roi_area: gpd.GeoDataFrame = backend.roi(ix=ix, context_area=context_area)
        roi_cells: gpd.GeoDataFrame = backend.roi_cells(
            roi_area=roi_area,
            cell_gdf=cell_gdf,
            predicate=predicate,
            silence_warnings=silence_warnings,
            parallel=parallel,
            num_processes=num_processes,
            **kwargs,
        )
        context_dict = {"roi_area": roi_area}
        context_dict["roi_cells"] = roi_cells

        # interface context
        iface_area: gpd.GeoDataFrame = backend.interface(
            top_roi_area=roi_area, bottom_gdf=context_area2, buffer_dist=buffer_dist
        )
        iface_cells: gpd.GeoDataFrame = backend.roi_cells(
            roi_area=iface_area,
            cell_gdf=cell_gdf,
            predicate=predicate,
            silence_warnings=silence_warnings,
            parallel=parallel,
            num_processes=num_processes,
            **kwargs,
        )
        context_dict["interface_area"] = iface_area
        context_dict["interface_cells"] = iface_cells

        # context networks
        if fit_network:
            if (iface_cells is None or iface_cells.empty) or (
                roi_cells is None or roi_cells.empty
            ):
                context_dict["full_network"] = None
                context_dict["roi_network"] = None
                context_dict["interface_network"] = None
                context_dict["border_network"] = None
            else:
                # merge the gdfs to compute union weights
                cells = pd.concat([roi_cells, iface_cells], sort=False)

                # fit the union graph
                context_dict["full_network"] = fit_graph(
                    cells,
                    type=graph_type,
                    id_col="global_id",
                    thresh=dist_thresh,
                    use_index=False,
                )

                # Get the weight subsets
                context_dict["roi_network"] = w_subset(
                    context_dict["full_network"],
                    sorted(set(roi_cells.global_id)),
                    silence_warnings=silence_warnings,
                )
                context_dict["interface_network"] = w_subset(
                    context_dict["full_network"],
                    sorted(set(iface_cells.global_id)),
                    silence_warnings=silence_warnings,
                )

                # get the weights 4 the nodes that have links crossing the iface border
                context_dict["border_network"] = get_border_crosser_links(
                    union_weights=context_dict["full_network"],
                    roi_weights=context_dict["roi_network"],
                    iface_weights=context_dict["interface_network"],
                    only_border_crossers=True,
                )

        if fit_grid:
            if grid_type == "hex":
                kwargs = {"resolution": resolution}
            else:
                kwargs = {
                    "patch_size": patch_size,
                    "stride": stride,
                    "pad": pad,
                    "predicate": predicate,
                }

            context_dict["roi_grid"] = fit_spatial_grid(
                gdf=roi_area, grid_type=grid_type, **kwargs
            )
            context_dict["interface_grid"] = fit_spatial_grid(
                gdf=iface_area, grid_type=grid_type, **kwargs
            )

        return context_dict

    def context2weights(self, key: str) -> W:
        """Merge the networks of type `key` into one spatial weights obj.

        Parameters:
            key (str):
                The key of the context dictionary that contains the spatial
                weights to be merged. One of "roi_network", "full_network",
                "interface_network", "border_network"

        Returns:
            libpysal.weights.W:
                A spatial weights object containing all the distinct networks
                in the context.
        """
        allowed = ("roi_network", "full_network", "interface_network", "border_network")
        if key not in allowed:
            raise ValueError(f"Illegal key. Got: {key}. Allowed: {allowed}")

        cxs = list(self.context.items())
        wout = W({0: [0]})
        for _, c in cxs:
            w = c[key]
            if isinstance(w, W):
                wout = w_union(wout, w, silence_warnings=True)

        # remove self loops
        wout = w_subset(wout, list(wout.neighbors.keys())[1:], silence_warnings=True)

        return wout

    def context2gdf(self, key: str) -> gpd.GeoDataFrame:
        """Merge the GeoDataFrames of type `key` into one geodataframe.

        Note:
            Returns None if no data is found.

        Parameters:
            key (str):
                The key of the context dictionary that contains the data to be converted
                to gdf. One of "roi_area", "roi_cells", "interface_area", "roi_grid",
                "interface_grid", "interface_cells", "roi_interface_cells"

        Returns:
            gpd.GeoDataFrame:
                Geo dataframe containing all the objects
        """
        allowed = (
            "roi_area",
            "roi_cells",
            "interface_area",
            "roi_grid",
            "interface_grid",
            "interface_cells",
            "roi_interface_cells",
        )
        if key not in allowed:
            raise ValueError(f"Illegal key. Got: {key}. Allowed: {allowed}")

        con = []
        for i in self.context.keys():
            if self.context[i][key] is not None:
                if isinstance(self.context[i][key], tuple):
                    con.append(self.context[i][key][0])
                else:
                    con.append(self.context[i][key])

        if not con:
            return

        gdf = pd.concat(
            con,
            keys=[i for i in self.context.keys() if self.context[i][key] is not None],
        )
        gdf = gdf.explode(ignore_index=True)

        return (
            gdf.reset_index(level=0, names="label")
            .drop_duplicates("geometry")
            .set_geometry("geometry")
        )

    def plot(
        self,
        key: str,
        network_key: str = None,
        grid_key: str = None,
        show_legends: bool = True,
        color: str = None,
        figsize: Tuple[int, int] = (12, 12),
        edge_kws: Dict[str, Any] = None,
        **kwargs,
    ) -> plt.Axes:
        """Plot the context with areas, cells, and interface areas highlighted.

        Parameters:
            key (str):
                The key of the context dictionary that contains the data to be plotted.
                One of "roi_area",
            network_key (str):
                The key of the context dictionary that contains the spatial weights to
                be plotted. One of "roi_network"
            grid_key (str):
                The key of the context dictionary that contains the grid to be plotted.
                One of "roi_grid"
            show_legends (bool):
                Flag, whether to include legends for each in the plot.
            color (str):
                A color for the interfaces or rois, Ignored if `show_legends=True`.
            figsize (Tuple[int, int]):
                Size of the figure.
            **kwargs (Dict[str, Any])]):
                Extra keyword arguments passed to the `plot` method of the
                geodataframes.

        Returns:
            AxesSubplot

        Examples:
            Plot the tumor-stroma areas.

            >>> from cellseg_gsontools.spatial_context import InterfaceContext
            >>> cells = read_gdf("cells.feather")
            >>> areas = read_gdf("areas.feather")
            >>> ts_iface = InterfaceContext(
            ...     area_gdf=areas,
            ...     cell_gdf=cells,
            ...     top_labels="tumor",
            ...     bottom_labels="stroma",
            ... )
            >>> ts_iface.fit(verbose=False)
            >>> ts_iface.plot("interface_area", show_legends=True)
            <AxesSubplot: >
        """
        allowed = ("roi_area", "interface_area")
        if key not in allowed:
            raise ValueError(f"Illegal key. Got: {key}. Allowed: {allowed}")

        context_gdf = self.context2gdf(key)

        grid_gdf = None
        if grid_key is not None:
            grid_gdf = self.context2gdf(grid_key)

        network_gdf = None
        if network_key is not None:
            edge_kws = edge_kws or {}
            w = self.context2weights(network_key)
            network_gdf = weights2gdf(self.cell_gdf, w)

        return plot_all(
            cell_gdf=self.cell_gdf.set_geometry("geometry"),
            area_gdf=self.area_gdf.set_geometry("geometry"),
            context_gdf=context_gdf,
            grid_gdf=grid_gdf,
            network_gdf=network_gdf,
            show_legends=show_legends,
            color=color,
            figsize=figsize,
            edge_kws=edge_kws,
            **kwargs,
        )



from typing import Callable, Optional, Tuple

import geopandas as gpd
import psutil
from pandarallel import pandarallel

__all__ = ["gdf_apply"]


def gdf_apply(
    gdf: gpd.GeoDataFrame,
    func: Callable,
    axis: int = 1,
    parallel: bool = True,
    num_processes: Optional[int] = -1,
    pbar: bool = False,
    columns: Optional[Tuple[str, ...]] = None,
    **kwargs,
) -> gpd.GeoSeries:
    """Apply or parallel apply a function to any col or row of a GeoDataFrame.

    Parameters:
        gdf (gpd.GeoDataFrame):
            Input GeoDataFrame.
        func (Callable):
            A callable function.
        axis (int, default=1):
            The gdf axis to apply the function on.axis=1 means rowise. axis=0
            means columnwise.
        parallel (bool, default=True):
            Flag, whether to parallelize the operation with `pandarallel`.
        num_processes (int, default=-1):
            The number of processes to use when parallel=True. If -1,
            this will use all available cores.
        pbar (bool, default=False):
            Show progress bar when executing in parallel mode. Ignored if
            `parallel=False`.
        columns (Optional[Tuple[str, ...]], default=None):
            A tuple of column names to apply the function on. If None,
            this will apply the function to all columns.
        **kwargs (Dict[str, Any]): Arbitrary keyword args for the `func` callable.

    Returns:
        gpd.GeoSeries:
            A GeoSeries object containing the computed values for each
            row or col in the input gdf.

    Examples:
        Get the compactness of the polygons in a gdf
        >>> from cellseg_gsontools import gdf_apply
        >>> gdf["compactness"] = gdf_apply(
        ...     gdf, compactness, columns=["geometry"], parallel=True
        ... )
    """
    if columns is not None:
        if not isinstance(columns, (tuple, list)):
            raise ValueError(f"columns must be a tuple or list, got {type(columns)}")
        gdf = gdf[columns]

    if not parallel:
        res = gdf.apply(lambda x: func(*x, **kwargs), axis=axis)
    else:
        cpus = psutil.cpu_count(logical=False) if num_processes == -1 else num_processes
        pandarallel.initialize(verbose=1, progress_bar=pbar, nb_workers=cpus)
        res = gdf.parallel_apply(lambda x: func(*x, **kwargs), axis=axis)

    return res



try:
    import dask_geopandas

    _has_dask_geopandas = True
except ImportError:
    _has_dask_geopandas = False


import geopandas as gpd

from cellseg_gsontools.spatial_context.backend import _SpatialBackend
from cellseg_gsontools.spatial_context.ops import (
    get_interface_zones,
    get_objs,
    get_objs_dgp,
)

__all__ = ["_SpatialContextDGP"]


class _SpatialContextDGP(_SpatialBackend):
    def __init__(self) -> None:
        """Create a dask_geopandas spatial context class."""
        if not _has_dask_geopandas:
            raise ImportError(
                "dask_geopandas is required to use the spatial context backend."
                "Install it with `pip install dask_geopandas`."
            )
        self.backend_name = "dask_geopandas"

    def roi(self, context_area: gpd.GeoDataFrame, ix: int) -> gpd.GeoDataFrame:
        """Get a roi area of index `ix`.

        Parameters
        ----------
        context_area : gpd.GeoDataFrame
            The context area gdf.
        ix : int
            The index of the roi area. I.e., the ith roi area.

        Returns
        -------
            gpd.GeoDataFrame:
                The ith roi area.
        """
        row: gpd.GeoSeries = context_area.loc[ix]
        roi_area = gpd.GeoDataFrame([row], crs=context_area.crs)

        return roi_area

    def roi_cells(
        self,
        roi_area: gpd.GeoDataFrame,
        cell_gdf: gpd.GeoDataFrame,
        cell_gdf_dgp: dask_geopandas.GeoDataFrame = None,
        predicate: str = "intersects",
        silence_warnings: bool = True,
        parallel: bool = False,
        **kwargs,
    ) -> gpd.GeoDataFrame:
        """Get the cells within the roi area.

        Parameters
        ----------
        roi_area : gpd.GeoDataFrame
            The roi area.
        cell_gdf : gpd.GeoDataFrame
            The cell gdf.
        cell_gdf_dgp : dask_geopandas.GeoDataFrame, optional
            The cell gdf as a dask_geopandas.GeoDataFrame. If None, then the geopandas
            cell_gdf is used.
        predicate : str, default="intersects"
            The predicate to use for the spatial join. See `geopandas.tools.sjoin`
        silence_warnings : bool, default=True
            Flag, whether to silence all the geopandas related warnings.
        parallel : bool, default=False
            Flag, whether to parallelize the spatial join.

        Returns
        -------
            gpd.GeoDataFrame:
                The cells within the roi area.

        Returns
        -------
            gpd.GeoDataFrame:
                The cells within the roi area.
        """
        if roi_area is None or roi_area.empty:
            return

        # get the cells within the roi area
        # if area is large, use dask-geopandas
        if (
            roi_area.geometry.area.values[0] > 1e8
            and parallel
            and cell_gdf_dgp is not None
        ):
            objs_within: gpd.GeoDataFrame = get_objs_dgp(
                roi_area,
                cell_gdf_dgp,
                silence_warnings=silence_warnings,
                predicate=predicate,
            )
        else:
            objs_within: gpd.GeoDataFrame = get_objs(
                roi_area,
                cell_gdf,
                silence_warnings=silence_warnings,
                predicate=predicate,
            )

        if objs_within is None or objs_within.empty:
            return

        # rename spatial join columns
        objs_within = objs_within.rename(
            columns={
                "index_right": "spatial_context_id",
                "global_id_left": "global_id",
                "class_name_left": "class_name",
                "class_name_right": "spatial_context",
            },
        )

        # drop unnecessary columns and return
        objs_within.drop(columns=["global_id_right"], inplace=True)

        # convert to geopandas if dask-geopandas
        if objs_within.index.name == "hilbert_distance":
            objs_within = self.to_geopandas(objs_within)

        return objs_within.set_geometry("geometry")

    def interface(
        self,
        top_roi_area: gpd.GeoDataFrame,
        bottom_gdf: gpd.GeoDataFrame,
        buffer_dist: int,
    ) -> gpd.GeoDataFrame:
        """Get an interface area of index `ix`.

        Parameters
        ----------
            top_roi_area : gpd.GeoDataFrame
                The top roi area.
            bottom_gdf : gpd.GeoDataFrame
                The bottom gdf on top of which top_roi_area is buffered.
        """
        # Get the intersection of the roi and the area of type `label2`
        iface = get_interface_zones(
            buffer_area=top_roi_area,
            areas=bottom_gdf,
            buffer_dist=buffer_dist,
        )

        # If there are many interfaces, dissolve them into one
        if len(iface) > 1:
            iface = iface.dissolve().set_geometry("geometry")

        return iface

    def to_geopandas(self, gdf):
        """Convert gdf to right SpatialContext gdf format."""
        return gdf.sort_values("global_id").set_index("global_id", drop=False)

    def convert_area_gdf(self, gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        """Convenience identity function to match the backend API.."""
        return gdf

    def convert_cell_gdf(
        self, gdf: gpd.GeoDataFrame, n_partitions: int = None, **kwargs
    ) -> dask_geopandas.GeoDataFrame:
        """Set up a dask_geopandas.GeoDataFrame attribute for parallel spatial ops.

        Parameters
        ----------
        gdf : gpd.GeoDataFrame
            The cell gdf to convert.
        n_partitions : int, optional
            The number of processes/partitions to use for the DaskGeoDataFrame. If None,
            the number of partitions is set to the number of cores.

        Attributes
        ----------
        cell_gdf_dgp : DaskGeoDataFrame
            The cell gdf as a dask_geopandas.GeoDataFrame.

        Returns
        -------
        gpd.GeoDataFrame:
            Return the original gpd.GeoDataFrame.
        """
        if _has_dask_geopandas:
            print(
                f"partitioning cell_gdf into {n_partitions} partitions and shuffling."
            )
            self.cell_gdf_dgp = dask_geopandas.from_geopandas(
                gdf, npartitions=n_partitions
            )
            self.cell_gdf_dgp = self.cell_gdf_dgp.spatial_shuffle()
        else:
            raise ImportError(
                "dask-geopandas is required to use the spatial context "
                "backend. with backend == 'dask-geopandas'."
                "Install with `pip install dask-geopandas`."
            )

        return gdf



from functools import partial
from typing import Optional, Sequence, Tuple, Union

import geopandas as gpd
import numpy as np
from libpysal.weights import W

from cellseg_gsontools.apply import gdf_apply
from cellseg_gsontools.neighbors import neighborhood, nhood_dists, nhood_vals
from cellseg_gsontools.utils import set_uid

__all__ = ["reduce", "local_character", "local_distances"]


def reduce(
    x: Sequence[Union[int, float]],
    areas: Optional[Sequence[float]] = None,
    how: str = "sum",
) -> float:
    """Reduce a numeric sequence.

    Note:
        Optionally can weight the input values based on area.

    Parameters:
        x (Sequence):
            The input value-vector. Shape (n, )
        areas (Sequence, optional):
            The areas of the spatial objects. This is for weighting. Optional.
        how (str, default="sum"):
            The reduction method for the neighborhood. One of:
            "sum", "mean", "median", "min", "max", "std".

    Raises:
        ValueError:
            If an illegal reduction method is given.

    Returns:
        float:
            The reduced value of the input array.
    """
    w = 1.0
    if areas is not None:
        w = areas / (np.sum(areas) + 1e-8)

    res = 0
    if how == "sum":
        res = np.sum(x * w)
    elif how == "mean":
        res = np.mean(x * w)
    elif how == "median":
        res = np.median(x * w)
    elif how == "max":
        res = np.max(x * w)
    elif how == "min":
        res = np.min(x * w)
    elif how == "std":
        res = np.std(x * w)
    else:
        allowed = ("sum", "mean", "median", "min", "max", "std")
        ValueError(f"Illegal param `how`. Got: {how}, Allowed: {allowed}")

    return float(res)


def local_character(
    gdf: gpd.GeoDataFrame,
    spatial_weights: W,
    val_col: Union[str, Tuple[str, ...]],
    id_col: str = None,
    reductions: Tuple[str, ...] = ("sum",),
    weight_by_area: bool = False,
    parallel: bool = True,
    rm_nhood_cols: bool = True,
    col_prefix: str = None,
    create_copy: bool = True,
) -> gpd.GeoDataFrame:
    """Compute the local sum/mean/median/min/max/std of a specified metric for each
    neighborhood of geometry objects in a gdf.

    Note:
        Option to weight the nhood values by their area before reductions.

    Parameters:
        gdf (gpd.GeoDataFrame):
            The input GeoDataFrame.
        spatial_weights (libysal.weights.W):
            Libpysal spatial weights object.
        val_col (Union[str, Tuple[str, ...]]):
            The name of the column in the gdf for which the reduction is computed.
            If a tuple, the reduction is computed for each column.
        id_col (str):
            The unique id column in the gdf. If None, this uses `set_uid` to set it.
            Defaults to None.
        reductions (Tuple[str, ...]):
            A list of reduction methods for the neighborhood. One of
            "sum", "mean", "median", "min", "max", "std". Defaults to ("sum", ).
        weight_by_area (bool):
            Flag whether to weight the neighborhood values by the area of the object.
            Defaults to False.
        parallel (bool):
            Flag whether to use parallel apply operations when computing the character.
            Defaults to True.
        rm_nhood_cols (bool):
            Flag, whether to remove the extra neighborhood columns from the result gdf.
            Defaults to True.
        col_prefix (str):
            Prefix for the new column names.
        create_copy (bool):
            Flag whether to create a copy of the input gdf and return that.
            Defaults to True.

    Returns:
        gpd.GeoDataFrame:
            The input geodataframe with computed character column added.

    Examples:
        Compute the mean of eccentricity values for each cell neighborhood
        >>> from cellseg_gsontools.character import local_character
        >>> from cellseg_gsontools.graphs import fit_graph
        >>> w = fit_graph(gdf, type="distband", thres=75.0)
        >>> local_character(
        ...     gdf,
        ...     spatial_weights=w,
        ...     val_col=["eccentricity", "area"],
        ...     reduction=["mean", "median"],
        ...     weight_by_area=True,
        ... )
    """
    allowed = ("sum", "mean", "median", "min", "max", "std")
    if not all(r in allowed for r in reductions):
        raise ValueError(
            f"Illegal reduction in `reductions`. Got: {reductions}. "
            f"Allowed reductions: {allowed}."
        )

    if create_copy:
        gdf = gdf.copy()

    # set uid
    if id_col is None:
        id_col = "uid"
        gdf = set_uid(gdf)

    # Get the immediate node neighborhood
    func = partial(neighborhood, spatial_weights=spatial_weights)
    gdf["nhood"] = gdf_apply(gdf, func, columns=[id_col], axis=1, parallel=parallel)

    # get areas
    area_col = None
    if weight_by_area:
        area_col = "nhood_areas"
        func = partial(nhood_vals, values=gdf.geometry.area)
        gdf[area_col] = gdf_apply(
            gdf, func, columns=["nhood"], axis=1, parallel=parallel
        )

    if isinstance(val_col, str):
        val_col = (val_col,)

    # get character values
    # Compute the neighborhood characters
    col_prefix = "" if col_prefix is None else col_prefix
    for col in val_col:
        values = gdf[col]
        char_col = f"{col}_nhood_vals"
        func = partial(nhood_vals, values=values)
        gdf[char_col] = gdf_apply(
            gdf,
            func,
            columns=["nhood"],
            axis=1,
            parallel=parallel,
        )

        # loop over the reduction methods
        for r in reductions:
            columns = [char_col]
            new_col = f"{col_prefix}{col}_nhood_{r}"
            if area_col in gdf.columns:
                columns.append(area_col)
                new_col = f"{col_prefix}{col}_nhood_{r}_area_weighted"

            func = partial(reduce, how=r)
            gdf[new_col] = gdf_apply(
                gdf,
                func,
                columns=columns,
                axis=1,
                parallel=parallel,
            )

        if rm_nhood_cols:
            gdf = gdf.drop(labels=[char_col], axis=1)

    if rm_nhood_cols:
        labs = ["nhood"]
        if weight_by_area:
            labs.append(area_col)
        gdf = gdf.drop(labels=labs, axis=1)

    return gdf


def local_distances(
    gdf: gpd.GeoDataFrame,
    spatial_weights: W,
    id_col: str = None,
    reductions: Tuple[str, ...] = ("mean",),
    weight_by_area: bool = False,
    invert: bool = False,
    parallel: bool = True,
    rm_nhood_cols: bool = True,
    col_prefix: str = None,
    create_copy: bool = True,
) -> gpd.GeoDataFrame:
    """Compute the local sum/mean/median/min/max/std distance of the neighborhood
    distances for each geometry object in a gdf.

    Note:
        Option to weight the nhood values by their area before reductions.

    Parameters:
        gdf (gpd.GeoDataFrame):
            The input GeoDataFrame.
        spatial_weights (libysal.weights.W):
            Libpysal spatial weights object.
        id_col (str):
            The unique id column in the gdf. If None, this uses `set_uid` to set it.
            Defaults to None.
        reductions (Tuple[str, ...]):
            A list of reduction methods for the neighborhood. One of "sum", "mean",
            "median", "min", "max", "std". Defaults to ("sum", ).
        weight_by_area (bool):
            Flag whether to weight the neighborhood values by the area of the object.
            Defaults to False.
        invert (bool):
            Flag whether to invert the distances. Defaults to False.
        parallel (bool):
            Flag whether to use parallel apply operations when computing the character.
            Defaults to True.
        rm_nhood_cols (bool):
            Flag, whether to remove the extra neighborhood columns from the result gdf.
            Defaults to True.
        col_prefix (str):
            Prefix for the new column names.
        create_copy (bool):
            Flag whether to create a copy of the input gdf and return that.
            Defaults to True.

    Returns:
        gpd.GeoDataFrame:
            The input geodataframe with computed distances column added.

    Examples:
        Compute the mean of eccentricity values for each neighborhood
        >>> from cellseg_gsontools.character import local_distances
        >>> from cellseg_gsontools.graphs import fit_graph
        >>> w = fit_graph(gdf, type="distband", thres=75.0)
        >>> local_distances(gdf, spatial_weights=w, reduction=["mean"], weight_by_area=True)
    """
    allowed = ("sum", "mean", "median", "min", "max", "std")
    if not all(r in allowed for r in reductions):
        raise ValueError(
            f"Illegal reduction in `reductions`. Got: {reductions}. "
            f"Allowed reductions: {allowed}."
        )

    if create_copy:
        gdf = gdf.copy()

    # set uid
    if id_col is None:
        id_col = "uid"
        gdf = set_uid(gdf)

    # get the immediate node neighborhood
    func = partial(neighborhood, spatial_weights=spatial_weights)
    gdf["nhood"] = gdf_apply(gdf, func, columns=[id_col], axis=1, parallel=parallel)

    # get areas
    area_col = None
    if weight_by_area:
        func = partial(nhood_vals, values=gdf.geometry.area)
        gdf[area_col] = gdf_apply(
            gdf, func, columns=["nhood"], axis=1, parallel=parallel
        )

    # get distances
    func = partial(nhood_dists, centroids=gdf.centroid, invert=invert)
    gdf["nhood_dists"] = gdf_apply(
        gdf, func, columns=["nhood"], axis=1, parallel=parallel
    )

    col_prefix = "" if col_prefix is None else col_prefix

    # loop over the reduction methods
    for r in reductions:
        for r in reductions:
            columns = ["nhood_dists"]
            new_col = f"{col_prefix}nhood_dists_{r}"
            if area_col in gdf.columns:
                columns.append(area_col)
                new_col = f"{col_prefix}nhood_dists_{r}_area_weighted"

            func = partial(reduce, how=r)
            gdf[new_col] = gdf_apply(
                gdf, func, columns=columns, axis=1, parallel=parallel
            )

    if rm_nhood_cols:
        labs = ["nhood"]
        if weight_by_area:
            labs.append(area_col)
        gdf = gdf.drop(labels=labs, axis=1)

    return gdf



from typing import Tuple, Union

import geopandas as gpd
import numpy as np
from libpysal.cg import alpha_shape_auto

from ..clustering import cluster_points
from .within import WithinContext

__all__ = ["PointClusterContext"]


class PointClusterContext(WithinContext):
    """Handle & extract dense point clusters from `cell_gdf`.

    Point-clusters are dense regions of points. This context is useful when you
    want to extract dense regions of points of type `label` from `cell_gdf` such as
    immune-cell clusters. The clusters are extracted using one of the clustering
    methods: "dbscan", "adbscan", "optics" after which the clusters are converted
    to polygons/areas using alpha-shapes.

    Note:
        This class inherits from `WithinContext` and thus has all the methods
        and attributes of that class.

    Note:
        `cell_gdf` has to contain a column named 'class_name'

    Parameters:
        cell_gdf (gpd.GeoDataFrame):
            A geo dataframe that contains small cellular objects that are
            enclosed by larger tissue areas in `area_gdf`.
        labels (Union[Tuple[str, ...], str]):
            The class name(s) of the areas of interest. The objects within
            these areas are extracted. E.g. "cancer" or "stroma".
        cluster_method (str):
            The clustering method used to extract the point-clusters. One of:
            "dbscan", "adbscan", "optics"
        min_area_size (float or str, optional):
            The minimum area of the objects that are kept. All the objects in
            the `area_gdf` that are larger are kept than `min_area_size`. If
            None, all the areas are kept. Defaults to None.
        graph_type (str):
            The type of the graph to be fitted to the cells inside interfaces.
            One of: "delaunay", "distband", "relative_nhood", "knn".
        dist_thresh (float):
            Distance threshold for the length of the network links.
        grid_type (str):
            The type of the grid to be fitted on the roi areas. One of:
            "square", "hex".
        patch_size (Tuple[int, int]):
            The size of the grid patches to be fitted on the context. This is
            used when `grid_type='square'`.
        stride (Tuple[int, int]):
            The stride of the sliding window for grid patching. This is used
            when `grid_type='square'`.
        pad (int):
            The padding to add to the bounding box on the grid. This is used
            when `grid_type='square'`.
        resolution (int):
            The resolution of the h3 hex grid. This is used when
            `grid_type='hex'`.
        predicate (str):
            The predicate to use for the spatial join when extracting the ROI
            cells. See `geopandas.tools.sjoin`
        silence_warnings (bool):
            Flag, whether to silence all the warnings.
        parallel (bool):
            Flag, whether to parallelize the context fitting. If
            `backend == "geopandas"`, the parallelization is implemented with
            `pandarallel` package. If `backend == "spatialpandas"`, or
            `backend == "dask-geopandas"` the parallelization is implemented
             with Dask library.
        num_processes (int):
            The number of processes to use when parallel=True. If -1, this
            will use all the available cores.
        backend (str):
            The backend to use for the spatial context. One of "geopandas",
            "spatialpandas" "dask-geopandas". "spatialpandas" or
            "dask-geopandas" is recommended for gdfs that may contain huge
            polygons.
        **kwargs (Dict[str, Any]):
            Arbitrary key-word arguments passed to the clustering methods.

    Attributes:
        context (Dict[int, Dict[str, Union[gpd.GeoDataFrame, libpysal.weights.W]]]):
            A nested dict that contains dicts for each of the distinct clusters.
            The keys of the outer dict are the indices of these areas. The inner dicts
            contain the keys:

            - `roi_area`- `gpd.GeoDataFrame`:  the alpha shape of a cell cluster.
            - `roi_cells` - `gpd.GeoDataFrame`: the cells that are contained
                    inside the `roi_area`.
            - `roi_network` - `libpysal.weights.W`: spatial weights network of
                    the cells inside the `roi_area`. This can be used to extract
                    graph features inside the clusters.
            - `roi_grid` - `gpd.GeoDataFrame`: of the grid fitted on the `roi_area`.
                    This can be used to extract grid features inside the clusters.

    Raises:
        ValueError:
            if `cell_gdf` don't contain 'class_name' column.

    Examples:
        Create a point cluster context and plot the cells inside one cluster area.
        >>> from cellseg_gsontools.spatial_context import ClusterContext
        >>> cell_gdf = pre_proc_gdf(read_gdf("cells.json"))
        >>> cluster_context = PointClusterContext(
        ...     cell_gdf=cell_gdf,
        ...     labels=["inflammatory"],
        ...     cluster_method="adbscan",
        ...     silence_warnings=True,
        ... )
        >>> cluster_context.fit(parallel=False, fit_graph=False)
        >>> cluster_context.plot("roi_area", show_legends=True)
        <AxesSubplot: >
    """

    def __init__(
        self,
        cell_gdf: gpd.GeoDataFrame,
        labels: Union[Tuple[str, ...], str],
        cluster_method: str = "dbscan",
        min_area_size: Union[float, str] = None,
        graph_type: str = "distband",
        dist_thresh: float = 100.0,
        grid_type: str = "square",
        patch_size: Tuple[int, int] = (256, 256),
        stride: Tuple[int, int] = (256, 256),
        pad: int = None,
        resolution: int = 9,
        predicate: str = "intersects",
        silence_warnings: bool = True,
        n_jobs: int = -1,
        parallel: bool = False,
        num_processes: int = -1,
        backend: str = "geopandas",
        **kwargs,
    ) -> None:
        cell_gdf.set_crs(epsg=4328, inplace=True, allow_override=True)
        area_gdf = self.run_clustering(
            cell_gdf, labels, cluster_method, n_jobs, **kwargs
        )

        if isinstance(labels, (list, tuple)):
            if len(labels) == 1:
                labels = labels[0]
            else:
                labels = "-".join(labels)

        super().__init__(
            area_gdf=area_gdf,
            cell_gdf=cell_gdf,
            labels=labels,
            min_area_size=min_area_size,
            dist_thresh=dist_thresh,
            graph_type=graph_type,
            patch_size=patch_size,
            stride=stride,
            pad=pad,
            predicate=predicate,
            silence_warnings=silence_warnings,
            parallel=parallel,
            num_processes=num_processes,
            backend=backend,
            grid_type=grid_type,
            resolution=resolution,
        )

    def run_clustering(
        self,
        cell_gdf: gpd.GeoDataFrame,
        labels: Union[Tuple[str, ...], str],
        cluster_method: str,
        n_jobs: int,
        **kwargs,
    ) -> gpd.GeoDataFrame:
        """Run clustering on the cells and convert the clusters to areas.

        Parameters:
            cell_gdf (gpd.GeoDataFrame):
                A geo dataframe that contains the cell/nuclei objects.
            labels (Union[Tuple[str, ...], str]):
                The class name(s of the objects of interest. E.g. "cancer", "immune".
            cluster_method (str):
                The clustering method used to extract the point-clusters. One of:
                "dbscan", "adbscan", "optics"
            n_jobs (int):
                The number of processes to use in clustering.
            **kwargs (Dict[str, Any]):
                Arbitrary key-word arguments passed to the clustering methods.

        Returns:
            gpd.GeoDataFrame:
                A gdf containing the areas of the clusters.
        """
        # cluster the cells
        if isinstance(labels, str):
            # a little faster than .isin
            cells = cell_gdf[cell_gdf["class_name"] == labels].copy()
        else:
            if len(labels) == 1:
                cells = cell_gdf[cell_gdf["class_name"] == labels[0]].copy()
            else:
                cells = cell_gdf[cell_gdf["class_name"].isin(labels)].copy()

        cells = cluster_points(cells, method=cluster_method, n_jobs=n_jobs, **kwargs)

        # convert the clusters to areas with alpha-shapes
        labs = cells["labels"].unique()
        area_data = {"geometry": []}
        for lab in labs:
            if lab == str(-1) or lab == int(-1):
                continue

            if isinstance(lab, str):
                lab = str(lab)

            c = cells[cells["labels"] == lab]
            coords = np.vstack([c.centroid.x, c.centroid.y]).T
            alpha_shape = alpha_shape_auto(coords, step=15)
            area_data["geometry"].append(alpha_shape.buffer(10.0).buffer(-20.0))

        area_gdf = gpd.GeoDataFrame(area_data)

        if isinstance(labels, (list, tuple)):
            if len(labels) == 1:
                labels = labels[0]
            else:
                labels = "-".join(labels)

        area_gdf["class_name"] = [labels] * len(area_gdf)

        return area_gdf
